from bloqade.ir.visitor import BloqadeIRTransformer
from bloqade.ir import scalar
from bloqade.ir.control import pulse, sequence, field, waveform


def is_literal(expr):
    return isinstance(expr, scalar.Literal)


def is_negative(expr):
    return isinstance(expr, scalar.Negative) or isinstance(expr, waveform.Negative)


def is_zero(expr):
    return isinstance(expr, scalar.Literal) and expr.value == 0


def is_one(expr):
    return isinstance(expr, scalar.Literal) and expr.value == 1


def is_scaled_waveform(expr):
    return isinstance(expr, waveform.Scale)


def is_constant_waveform(expr):
    return isinstance(expr, waveform.Constant)


class Canonicalizer(BloqadeIRTransformer):
    def minmax_canonicalize(self, op, exprs):
        new_exprs = set()
        new_literals = set()

        for expr in exprs:
            if isinstance(expr, op):
                exprs = list(map(self.visit, expr.exprs))
                new_exprs.update(expr for expr in exprs if not is_literal(expr))
                new_literals.update(expr for expr in exprs if is_literal(expr))
            else:
                expr = self.visit(expr)
                if is_literal(expr):
                    new_literals.add(expr)
                else:
                    new_exprs.add(expr)

        if new_literals:
            if len(new_literals) > 1:
                minmax = min if op == scalar.Min else max
                new_literal = scalar.Literal(
                    minmax(*[ele.value for ele in new_literals])
                )
            else:
                new_literal = new_literals.pop()

            new_exprs.add(new_literal)

        if len(new_exprs) > 1:
            return op(exprs=frozenset(new_exprs))
        else:
            (new_expr,) = new_exprs
            return new_expr

    def visit_scalar_Min(self, node: scalar.Min):
        return self.minmax_canonicalize(scalar.Min, node.exprs)

    def visit_scalar_Max(self, node: scalar.Max):
        return self.minmax_canonicalize(scalar.Max, node.exprs)

    def visit_scalar_Negative(self, node: scalar.Negative):
        expr = self.visit(node.expr)
        if is_negative(expr):
            return expr.expr
        elif is_literal(expr):
            return scalar.Literal(-expr.value)
        else:  # only apply negative to literals
            return scalar.Negative(expr=expr)

    def visit_scalar_Add(self, node: scalar.Add):
        lhs = self.visit(node.lhs)
        rhs = self.visit(node.rhs)

        if lhs == rhs:
            return 2 * lhs
        elif is_zero(rhs):
            return lhs
        elif is_zero(lhs):
            return rhs
        elif is_literal(lhs) and is_literal(rhs):
            return scalar.Literal(lhs.value + rhs.value)
        elif is_negative(lhs) and is_negative(rhs):
            return self.visit(
                scalar.Negative(expr=scalar.Add(lhs=lhs.expr, rhs=rhs.expr))
            )
        elif (is_negative(lhs) and lhs.expr == rhs) or (
            is_negative(rhs) and rhs.expr == lhs
        ):
            return scalar.Literal(0)

        else:
            return scalar.Add(lhs=lhs, rhs=rhs)

    def visit_scalar_Mul(self, node: scalar.Mul):
        lhs = self.visit(node.lhs)
        rhs = self.visit(node.rhs)

        if is_zero(lhs) or is_zero(rhs):
            return scalar.Literal(0)
        elif is_one(lhs):
            return rhs
        elif is_one(rhs):
            return lhs
        elif is_literal(lhs) and is_literal(rhs):
            return scalar.Literal(lhs.value * rhs.value)
        elif is_negative(lhs) and is_negative(rhs):
            return self.visit(scalar.Mul(lhs=lhs.expr, rhs=rhs.expr))
        elif is_negative(lhs):
            return self.visit(scalar.Negative(expr=scalar.Mul(lhs=lhs.expr, rhs=rhs)))
        elif is_negative(rhs):
            return self.visit(scalar.Negative(expr=scalar.Mul(lhs=lhs, rhs=rhs.expr)))
        else:
            return scalar.Mul(lhs=lhs, rhs=rhs)

    def visit_scalar_Div(self, node: scalar.Div):
        lhs = self.visit(node.lhs)
        rhs = self.visit(node.rhs)

        if is_zero(lhs):
            return scalar.Literal(0)
        elif is_literal(lhs) and is_literal(rhs):
            return scalar.Literal(lhs.value / rhs.value)
        elif is_literal(rhs) and rhs.value == 1:
            return lhs
        elif is_negative(lhs) and is_negative(rhs):
            return self.visit(scalar.Div(lhs=lhs.expr, rhs=rhs.expr))
        elif is_negative(lhs):
            return self.visit(
                scalar.Negative(expr=self.visit(scalar.Div(lhs=lhs.expr, rhs=rhs)))
            )
        elif is_negative(rhs):
            return self.visit(
                scalar.Negative(expr=self.visit(scalar.Div(lhs=lhs, rhs=rhs.expr)))
            )
        else:
            return scalar.Div(lhs=lhs, rhs=rhs)

    ########################################
    #    Waveform Canonicalization Pass    #
    ########################################

    def visit_waveform_Append(self, node: waveform.Append):
        waveforms_pass_one = []

        # flatten nested append nodes
        for sub_waveform in map(self.visit, node.waveforms):
            if sub_waveform.duration == scalar.Literal(0):
                continue
            elif isinstance(sub_waveform, waveform.Append):
                waveforms_pass_one.extend(sub_waveform.waveforms)
            else:
                waveforms_pass_one.append(sub_waveform)

        # combine adjacent constant waveforms if possible
        waveforms_pass_two = [waveforms_pass_one[0]]
        for next_wf in waveforms_pass_one[1:]:
            last_wf = waveforms_pass_two.pop()
            if (
                is_constant_waveform(last_wf)
                and is_constant_waveform(next_wf)
                and last_wf.value == next_wf.value
            ):
                new_duration = self.visit(last_wf.duration + next_wf.duration)
                waveforms_pass_two.append(
                    waveform.Constant(value=last_wf.value, duration=new_duration)
                )
            else:
                waveforms_pass_two.append(last_wf)
                waveforms_pass_two.append(next_wf)

        # if there is only one waveform, return it
        if len(waveforms_pass_two) == 1:
            return waveforms_pass_two[0]
        else:
            return waveform.Append(waveforms=waveforms_pass_two)

    def visit_waveform_Add(self, node: waveform.Add):
        left = self.visit(node.left)
        right = self.visit(node.right)

        if left == right:
            return self.visit(waveform.Scale(2, waveform=left))
        elif is_zero(left.duration) or (
            is_constant_waveform(left) and is_zero(left.value)
        ):
            return right
        elif is_zero(right.duration) or (
            is_constant_waveform(right) and is_zero(right.value)
        ):
            return left
        elif (
            is_constant_waveform(left)
            and is_constant_waveform(right)
            and (left.duration == right.duration)
        ):
            return self.visit(
                waveform.Constant(
                    value=left.value + right.value, duration=left.duration
                )
            )
        elif (
            is_scaled_waveform(left)
            and is_scaled_waveform(right)
            and left.scalar == right.scalar
        ):
            new_waveform = left.waveform + right.waveform
            return self.visit(waveform.Scale(scalar=left.scalar, waveform=new_waveform))
        elif (
            is_scaled_waveform(left)
            and is_scaled_waveform(right)
            and left.waveform == right.waveform
        ):
            new_scalar = left.scalar + right.scalar
            return self.visit(waveform.Scale(scalar=new_scalar, waveform=left.waveform))
        elif is_scaled_waveform(left) and left.waveform == right:
            return self.visit(
                waveform.Scale(scalar=left.scalar + 1, waveform=left.waveform)
            )
        elif is_scaled_waveform(right) and right.waveform == left:
            return self.visit(
                waveform.Scale(scalar=right.scalar + 1, waveform=right.waveform)
            )
        else:
            return waveform.Add(left=left, right=right)

    def visit_waveform_Scale(self, node: waveform.Scale):
        scale = self.visit(node.scalar)
        sub_waveform = self.visit(node.waveform)

        if is_zero(scale):
            return waveform.Constant(0, duration=sub_waveform.duration)
        elif is_one(scale):
            return sub_waveform
        elif is_scaled_waveform(sub_waveform):
            return waveform.Scale(
                scalar=scale * sub_waveform.scalar, waveform=sub_waveform.waveform
            )

        elif is_constant_waveform(sub_waveform):
            return waveform.Constant(
                value=scale * sub_waveform.value, duration=sub_waveform.duration
            )
        else:
            return waveform.Scale(scalar=scale, waveform=sub_waveform)

    def visit_waveform_Slice(self, node: waveform.Slice):
        sub_waveform = self.visit(node.waveform)
        interval = self.visit(node.interval)

        if (
            interval.start == scalar.Literal(0)
            and interval.stop == sub_waveform.duration
        ):
            return sub_waveform
        elif is_scaled_waveform(sub_waveform):
            new_waveform = waveform.Slice(
                waveform=sub_waveform.waveform, interval=interval
            )
            return waveform.Scale(
                scalar=sub_waveform.scalar,
                waveform=new_waveform,
            )

        elif is_negative(sub_waveform):
            return waveform.Negative(
                waveform.Slice(waveform=sub_waveform.waveform, interval=interval)
            )

        else:
            return waveform.Slice(waveform=sub_waveform, interval=interval)

    def visit_waveform_Negative(self, node: waveform.Negative):
        sub_waveform = self.visit(node.waveform)

        if isinstance(sub_waveform, waveform.Negative):
            return sub_waveform.waveform
        elif is_constant_waveform(sub_waveform):
            new_value = -sub_waveform.value
            return waveform.Constant(
                value=new_value,
                duration=sub_waveform.duration,
            )
        elif is_scaled_waveform(sub_waveform) and is_negative(sub_waveform.scalar):
            new_scalar = -sub_waveform.scalar
            return waveform.Scale(
                scalar=new_scalar,
                waveform=sub_waveform.waveform,
            )
        else:
            return waveform.Negative(waveform=sub_waveform)

    ########################################
    #    Field Canonicalization Pass       #
    ########################################

    def visit_field_Field(self, node: field.Field):
        inv_drives = {}

        # map spatial modulations to waveforms
        for sm, wf in node.drives.items():
            wf = self.visit(wf)
            inv_drives[wf] = inv_drives.get(wf, []) + [sm]

        new_inv_drives = {}
        # merge spatial modulations with the same waveform
        for wf, sms in inv_drives.items():
            other_sms = []
            new_scaled_locations = {}

            for sm in sms:  # merge scaled locations
                if isinstance(sm, field.ScaledLocations):
                    for loc, scale in sm.value.items():
                        new_scaled_locations[loc] = (
                            new_scaled_locations.get(loc, scalar.Literal(0)) + scale
                        )
                else:
                    other_sms.append(sm)

            if new_scaled_locations:
                # if there are any scaled locations,
                # add them to the list of spatial modulations
                other_sms.append(field.ScaledLocations.create(new_scaled_locations))

            new_inv_drives[wf] = other_sms

        drives = {sm: wf for wf, sms in new_inv_drives.items() for sm in sms}

        return field.Field(drives=drives)

    ########################################
    #    Pulse Canonicalization Pass       #
    ########################################

    def visit_pulse_Slice(self, node: pulse.Slice):
        sub_pulse = self.visit(node.pulse)
        interval = self.visit(node.interval)

        if interval.start == scalar.Literal(0) and interval.stop == sub_pulse.duration:
            return sub_pulse
        else:
            return pulse.Slice(pulse=sub_pulse, interval=interval)

    def visit_pulse_Append(self, node: pulse.Append):
        pulses = []
        for p in map(self.visit, node.pulses):
            if p.duration == scalar.Literal(0):
                continue
            elif isinstance(p, pulse.Append):
                pulses.extend(p.pulses)
            else:
                pulses.append(p)

        if len(pulses) == 1:
            return pulses[0]
        else:
            return pulse.Append(pulses=pulses)

    ########################################
    #    Sequence Canonicalization Pass    #
    ########################################

    def visit_sequence_Slice(self, node: sequence.Slice):
        sub_sequence = self.visit(node.sequence)
        interval = self.visit(node.interval)

        if (
            interval.start == scalar.Literal(0)
            and interval.stop == sub_sequence.duration
        ):
            return sub_sequence
        else:
            return sequence.Slice(sequence=sub_sequence, interval=interval)

    def visit_sequence_Append(self, node: sequence.Append):
        sequences = []

        for sub_sequence in map(self.visit, node.sequences):
            if sub_sequence.duration == scalar.Literal(0):
                continue
            elif isinstance(sub_sequence, sequence.Append):
                sequences.extend(sub_sequence.sequences)
            else:
                sequences.append(sub_sequence)

        if len(sequences) == 1:
            return sequences[0]
        else:
            return sequence.Append(sequences=sequences)



from bloqade import start
from bloqade.compiler.codegen.python.emulator_ir import EmulatorProgramCodeGen
from bloqade.emulate.codegen.hamiltonian import RydbergHamiltonianCodeGen
from bloqade.emulate.ir.state_vector import AnalogGate
import numpy as np
import pytest


def callback(register, *_):
    return register.data


def test_run_callback():
    program = (
        start.add_position((0, 0))
        .add_position((0, 6.1))
        .rydberg.detuning.uniform.constant(15, 0.1)
        .rabi.amplitude.uniform.constant(15, 0.1)
    )

    bloqade_ir = program.parse_circuit()
    emulatir_ir = EmulatorProgramCodeGen().emit(bloqade_ir)
    hamiltonian = RydbergHamiltonianCodeGen().emit(emulatir_ir)

    state = hamiltonian.space.zero_state(np.complex128)
    (expected_result,) = AnalogGate(hamiltonian).apply(state)

    (result_single,) = program.bloqade.python().run_callback(callback)
    (result_multi,) = program.bloqade.python().run_callback(
        callback, multiprocessing=True, num_workers=1
    )

    np.testing.assert_equal(expected_result, result_single)
    np.testing.assert_equal(expected_result, result_multi)


def callback_exception(*args):
    raise ValueError


def test_run_callback_exception():
    program = (
        start.add_position((0, 0))
        .add_position((0, 6.1))
        .rydberg.detuning.uniform.constant(15, 1)
        .rabi.amplitude.uniform.constant(15, 1)
    )

    # program.bloqade.python().run_callback(callback_exception)

    with pytest.raises(RuntimeError):
        program.bloqade.python().run_callback(callback_exception)

    with pytest.raises(RuntimeError):
        program.bloqade.python().run_callback(
            callback_exception, multiprocessing=True, num_workers=1
        )

    (result_single,) = program.bloqade.python().run_callback(
        callback_exception, ignore_exceptions=True
    )
    (result_multi,) = program.bloqade.python().run_callback(
        callback_exception, multiprocessing=True, num_workers=1, ignore_exceptions=True
    )

    assert isinstance(result_single, ValueError)
    assert isinstance(result_multi, ValueError)


if __name__ == "__main__":
    test_run_callback()
    test_run_callback_exception()



from .define import (
    analyze_channels,
    canonicalize_circuit,
    assign_circuit,
    validate_waveforms,
    generate_ahs_code,
    generate_quera_ir,
    generate_braket_ir,
)

__all__ = [
    "analyze_channels",
    "canonicalize_circuit",
    "assign_circuit",
    "validate_waveforms",
    "generate_ahs_code",
    "generate_quera_ir",
    "generate_braket_ir",
]



from functools import cached_property
from dataclasses import fields


class HashTrait:
    @cached_property
    def _hash_value(self):
        value = hash(self.__class__)
        for field in fields(self):
            field_value = getattr(self, field.name)
            if isinstance(field_value, list):
                value ^= hash(tuple(field_value))
            elif isinstance(field_value, dict):
                value ^= hash(frozenset(field_value.items()))
            else:
                value ^= hash(field_value)

        return value

    def __hash__(self):
        return self._hash_value



from decimal import Decimal

import pytest
from bloqade import cast
from bloqade.ir.location import ListOfLocations, ParallelRegister
from bloqade.compiler.codegen.hardware.lattice import GenerateLattice
from bloqade.submission.capabilities import get_capabilities
from bloqade.submission.ir.parallel import ParallelDecoder, ClusterLocationInfo
from bloqade.ir.analog_circuit import AnalogCircuit
from bloqade.ir.control.sequence import Sequence


def test():
    lattice = ListOfLocations().add_position((0, 0)).add_position((0, 6), filling=False)

    capabilities = get_capabilities()

    capabilities.capabilities.lattice.area.height = Decimal("13.0e-6")
    capabilities.capabilities.lattice.area.width = Decimal("13.0e-6")
    capabilities.capabilities.lattice.geometry.number_sites_max = 4

    circuit = AnalogCircuit(lattice, Sequence({}))

    ahs_lattice_data = GenerateLattice(capabilities).emit(circuit)

    assert ahs_lattice_data.sites == [
        (Decimal("0.0"), Decimal("0.0")),
        (Decimal("0.0"), Decimal("6.0")),
    ]
    assert ahs_lattice_data.filling == [1, 0]
    assert ahs_lattice_data.parallel_decoder is None

    parallel_lattice = ParallelRegister(lattice, cast(5))

    with pytest.raises(ValueError):
        ahs_lattice_data = GenerateLattice().emit(parallel_lattice)

    ahs_lattice_data = GenerateLattice(capabilities).emit(parallel_lattice)

    assert ahs_lattice_data.sites == [
        (Decimal("0.0"), Decimal("0.0")),
        (Decimal("0.0"), Decimal("6.0")),
        (Decimal("5.0"), Decimal("0.0")),
        (Decimal("5.0"), Decimal("6.0")),
    ]
    assert ahs_lattice_data.filling == [1, 0, 1, 0]
    assert ahs_lattice_data.parallel_decoder == ParallelDecoder(
        [
            ClusterLocationInfo(
                cluster_index=(0, 0), global_location_index=0, cluster_location_index=0
            ),
            ClusterLocationInfo(
                cluster_index=(0, 0), global_location_index=1, cluster_location_index=1
            ),
            ClusterLocationInfo(
                cluster_index=(1, 0), global_location_index=2, cluster_location_index=0
            ),
            ClusterLocationInfo(
                cluster_index=(1, 0), global_location_index=3, cluster_location_index=1
            ),
        ]
    )



import plum
from bloqade.emulate.ir.emulator import EmulatorProgram
from bloqade.emulate.ir.space import Space
from bloqade.emulate.sparse_operator import (
    IndexMapping,
    SparseMatrixCSC,
    SparseMatrixCSR,
)
from dataclasses import dataclass, field
from numpy.typing import NDArray
from beartype.typing import List, Callable, Union, Optional, Tuple
from beartype.vale import IsAttr, IsEqual
from typing import Annotated
from beartype import beartype
import numpy as np
from scipy.integrate import ode
from numba import njit

SparseOperator = Union[IndexMapping, SparseMatrixCSR, SparseMatrixCSC]


RealArray = Annotated[NDArray[np.floating], IsAttr["ndim", IsEqual[1]]]
Complexarray = Annotated[NDArray[np.complexfloating], IsAttr["ndim", IsEqual[1]]]
StateArray = Union[RealArray, Complexarray]


@njit(cache=True)
def _expt_one_body_op(configs, n_level, psi, site, op):
    res = np.zeros(psi.shape[1:], dtype=np.complex128)

    divisor = n_level**site

    for i, config in enumerate(configs):
        col = (config // divisor) % n_level
        for row, ele in enumerate(op[:, col]):
            new_config = config - (col * divisor) + (row * divisor)

            j = np.searchsorted(configs, new_config)

            if j < configs.size and configs[j] == new_config:
                res += ele * psi[i, ...] * np.conj(psi[j, ...])

    return res


@njit(cache=True)
def _expt_two_body_op(configs, n_level, psi, sites, data, indices, indptr):
    res = np.zeros(psi.shape[1:], dtype=np.complex128)

    divisor_1 = n_level ** sites[1]
    divisor_2 = n_level ** sites[0]

    for i, config in enumerate(configs):
        col_1 = (config // divisor_1) % n_level
        col_2 = (config // divisor_2) % n_level
        col = col_1 + (col_2 * n_level)

        start = indptr[col]
        end = indptr[col + 1]

        for ele, row in zip(data[start:end], indices[start:end]):
            row_2, row_1 = divmod(row, n_level)

            new_config = (
                config
                - (col_1 * divisor_1)
                + (row_1 * divisor_1)
                - (col_2 * divisor_2)
                + (row_2 * divisor_2)
            )

            j = np.searchsorted(configs, new_config)

            if j < configs.size and configs[j] == new_config:
                res += ele * psi[i, ...] * np.conj(psi[j, ...])

    return res


@dataclass(frozen=True)
class StateVector:
    data: NDArray
    space: Space

    @plum.dispatch
    def _trace(self, matrix: np.ndarray, site_indices: Tuple[int, int]) -> complex:
        from scipy.sparse import csc_array

        shape = (self.space.atom_type.n_level**2, self.space.atom_type.n_level**2)

        if matrix.shape != shape:
            raise ValueError(
                f"expecting operator to be size {shape}, got site {matrix.shape}"
            )

        csc = csc_array(matrix)

        value = _expt_two_body_op(
            configs=self.space.configurations,
            n_level=self.space.atom_type.n_level,
            psi=self.data,
            sites=site_indices,
            data=csc.data,
            indices=csc.indices,
            indptr=csc.indptr,
        )

        return complex(value.real, value.imag)

    @plum.dispatch
    def _trace(self, matrix: np.ndarray, site_index: int) -> complex:
        shape = (self.space.atom_type.n_level, self.space.atom_type.n_level)

        if matrix.shape != shape:
            raise ValueError(
                f"expecting operator to be size {shape}, got {matrix.shape}"
            )

        value = _expt_one_body_op(
            configs=self.space.configurations,
            n_level=self.space.atom_type.n_level,
            psi=self.data,
            site=site_index,
            op=matrix,
        )

        return complex(value.real, value.imag)

    def local_trace(
        self, matrix: np.ndarray, site_indices: Union[int, Tuple[int, int]]
    ) -> complex:
        """return trace of an operator over the StateVector.

        Args:
            matrix (np.ndarray): Square matrix representing operator in the local
                hilbert space.
            site_indices (Union[int, Tuple[int, int]]): sites to apply operator to.

        Returns:
            complex: the trace of the operator over the state-vector.
        """
        return self._trace(matrix, site_indices)


@dataclass(frozen=True)
class DetuningOperator:
    diagonal: NDArray
    amplitude: Optional[Callable[[float], float]] = None

    def get_diagonal(self, time: float):
        if self.amplitude:
            return self.diagonal * self.amplitude(time)
        else:
            return self.diagonal


@dataclass(frozen=True)
class RabiOperator:
    op: SparseOperator
    amplitude: Callable[[float], float]
    phase: Optional[Callable[[float], float]] = None

    def dot(self, register: NDArray, output: NDArray, time: float):
        amplitude = self.amplitude(time) / 2
        if self.phase is None:
            return self.op.matvec(register, out=output, scale=amplitude)

        amplitude *= np.exp(1j * self.phase(time))
        self.op.matvec(register, out=output, scale=amplitude)
        self.op.T.matvec(register, out=output, scale=np.conj(amplitude))

        return output


@dataclass(frozen=True)
class RydbergHamiltonian:
    emulator_ir: EmulatorProgram
    space: Space
    rydberg: NDArray
    detuning_ops: List[DetuningOperator] = field(default_factory=list)
    rabi_ops: List[RabiOperator] = field(default_factory=list)

    def _ode_complex_kernel(self, time: float, register: NDArray, output: NDArray):
        diagonal = sum(
            (detuning.get_diagonal(time) for detuning in self.detuning_ops),
            start=self.rydberg,
        )

        np.multiply(diagonal, register, out=output)
        for rabi_op in self.rabi_ops:
            rabi_op.dot(register, output, time)

        output *= -1j
        return output

    def _ode_real_kernel(self, time: float, register: NDArray, output: NDArray):
        # this is needed to use solver that only work on real-valued states
        return self._ode_complex_kernel(
            time, register.view(np.complex128), output
        ).view(np.float64)

    def _ode_complex_kernel_int(self, time: float, register: NDArray, output: NDArray):
        diagonal = sum(
            (detuning.get_diagonal(time) for detuning in self.detuning_ops),
        )

        u = np.exp(-1j * self.rydberg * time)

        int_register = u * register

        np.multiply(diagonal, int_register, out=output)
        for rabi_op in self.rabi_ops:
            rabi_op.dot(int_register, output, time)

        np.conj(u, out=u)
        np.multiply(u, output, out=output)

        output *= -1j
        return output

    def _ode_real_kernel_int(self, time: float, register: NDArray, output: NDArray):
        # this is needed to use solver that only work on real-valued states
        return self._ode_complex_kernel_int(
            time, register.view(np.complex128), output
        ).view(np.float64)

    def _check_register(self, register: np.ndarray):
        register_shape = (self.space.size,)
        if register.shape != register_shape:
            raise ValueError(
                f"Expecting `register` to have  shape {register_shape}, "
                f"got shape {register.shape}"
            )

    def _apply(
        self,
        register_data: np.ndarray,
        time: Optional[float] = None,
        output: Optional[NDArray] = None,
    ) -> np.ndarray:
        self._check_register(register_data)

        if time is None:
            time = self.emulator_ir.duration

        if output is None:
            output = np.zeros_like(register_data, dtype=np.complex128)

        diagonal = sum(
            (detuning.get_diagonal(time) for detuning in self.detuning_ops),
        )

        np.multiply(diagonal, register_data, out=output)

        for rabi_op in self.rabi_ops:
            rabi_op.dot(register_data, output, time)

        return output

    @beartype
    def average(
        self,
        register: StateVector,
        time: Optional[float] = None,
    ) -> float:
        """Get energy average from RydbergHamiltonian object at time `time` with
        register `register`

        Args:
            register (StateVector): The state vector to take average with
            time (Optional[float], optional): Time value to evaluate average at.
            Defaults to duration of RydbergHamiltonian.

        Returns:
            float: average energy at time `time`
        """
        return np.vdot(register.data, self._apply(register.data, time)).real

    @beartype
    def average_and_variance(
        self,
        register: StateVector,
        time: Optional[float] = None,
    ) -> Tuple[float, float]:
        """Get energy average and variance from RydbergHamiltonian object at time `time`
        with register `register`

        Args:
            register (StateVector): The state vector to take average and variance with
            time (Optional[float], optional): Time value to evaluate average at.
            Defaults to duration of RydbergHamiltonian.

        Returns:
            Tuple[float, float]: average and variance of energy at time `time`
            respectively.
        """
        H_register_data = self._apply(register.data, time)

        average = np.vdot(register.data, H_register_data).real
        square_average = np.vdot(H_register_data, H_register_data).real

        return average, square_average - average**2

    @beartype
    def variance(
        self,
        register: StateVector,
        time: Optional[float] = None,
    ) -> float:
        """Get the energy variance from RydbergHamiltonian object at
        time `time` with register `register`

        Args:
            register (StateVector): The state vector to take variance with
            time (Optional[float], optional): Time value to evaluate average at.
            Defaults to duration of RydbergHamiltonian.

        Returns:
            complex: variance of energy at time `time` respectively.
        """

        _, var = self.average_and_variance(register, time)
        return var

    @plum.dispatch
    def expectation_value(  # noqa: F811
        self, register: np.ndarray, operator: np.ndarray, site_indices: int
    ) -> complex:
        """Calculate expectation values of one and two body operators.

        Args:
            register (np.ndarray): Register to evaluate expectation value with
            operator (np.ndarray): Operator to take expectation value of.
            site_indices (int, Tuple[int, int]): site/sites to evaluate `operator` at.
                It can either a single integer or a tuple of two integers for one and
                two body operator respectively.

        Raises:
            ValueError: Error is raised when the dimension of `operator` is not
            consistent with `site` argument. The size of the operator must fit the
            size of the local hilbert space of `site` depending on the number of sites
            and the number of levels inside each atom, e.g. for two site expectation v
            alue with a three level atom the operator must be a 9 by 9 array.

        Returns:
            complex: The expectation value.
        """
        self._check_register(register)


@dataclass(frozen=True)
class AnalogGate:
    SUPPORTED_SOLVERS = ["lsoda", "dop853", "dopri5"]

    hamiltonian: RydbergHamiltonian

    @staticmethod
    def _error_check_dop(status_code: int):
        if status_code in [1, 2]:
            return
        elif status_code == -1:
            raise RuntimeError("DOP853/DOPRI5: Input is not consistent.")
        elif status_code == -2:
            raise RuntimeError("DOP853/DOPRI5: Larger nsteps is needed.")
        elif status_code == -3:
            raise RuntimeError("DOP853/DOPRI5: Step size becomes too small.")
        elif status_code == -4:
            raise RuntimeError(
                "DOP853/DOPRI5: Problem is probably stiff (interrupted)."
            )
        else:
            raise RuntimeError(f"DOP853/DOPRI5: unhandled status code {status_code}")

    @staticmethod
    def _error_check_lsoda(status_code: int):
        if status_code == 2:
            return
        elif status_code == -1:
            raise RuntimeError(
                "LSODA: Excess work done on this call (perhaps wrong Dfun type)."
            )
        elif status_code == -2:
            raise RuntimeError(
                "LSODA: Excess accuracy requested (tolerances too small)."
            )
        elif status_code == -3:
            raise RuntimeError("LSODA: Illegal input detected (internal error).")
        elif status_code == -4:
            raise RuntimeError("LSODA: Repeated error test failures (internal error).")
        elif status_code == -5:
            raise RuntimeError(
                "LSODA: Repeated convergence failures "
                "(perhaps bad Jacobian or tolerances)."
            )
        elif status_code == -6:
            raise RuntimeError("LSODA: Error weight became zero during problem.")
        elif status_code == -7:
            raise RuntimeError(
                "LSODA: Internal workspace insufficient to finish (internal error)."
            )
        else:
            raise RuntimeError(f"LSODA: unhandled status code {status_code}")

    @staticmethod
    def _error_check(solver_name: str, status_code: int):
        if solver_name == "lsoda":
            AnalogGate._error_check_lsoda(status_code)
        elif solver_name in ["dop853", "dopri5"]:
            AnalogGate._error_check_dop(status_code)

    def _apply(
        self,
        state: StateArray,
        solver_name: str = "dop853",
        atol: float = 1e-7,
        rtol: float = 1e-14,
        nsteps: int = 2_147_483_647,
        times: Union[List[float], RealArray] = [],
    ):
        if state is None:
            state = self.hamiltonian.space.zero_state()

        if solver_name not in AnalogGate.SUPPORTED_SOLVERS:
            raise ValueError(f"'{solver_name}' not supported.")

        duration = self.hamiltonian.emulator_ir.duration

        state = np.asarray(state).astype(np.complex128, copy=False)

        solver = ode(self.hamiltonian._ode_real_kernel)
        solver.set_f_params(np.zeros_like(state, dtype=np.complex128))
        solver.set_initial_value(state.view(np.float64))
        solver.set_integrator(solver_name, atol=atol, rtol=rtol, nsteps=nsteps)

        if any(time >= duration or time < 0.0 for time in times):
            raise ValueError("Times must be between 0 and duration.")

        times = [*times, duration]

        for time in times:
            if time == 0.0:
                yield state
                continue
            solver.integrate(time)
            AnalogGate._error_check(solver_name, solver.get_return_code())
            yield solver.y.view(np.complex128)

    def _apply_interation_picture(
        self,
        state: StateArray,
        solver_name: str = "dop853",
        atol: float = 1e-7,
        rtol: float = 1e-14,
        nsteps: int = 2_147_483_647,
        times: Union[List[float], RealArray] = [],
    ):
        if state is None:
            state = self.hamiltonian.space.zero_state()

        if solver_name not in AnalogGate.SUPPORTED_SOLVERS:
            raise ValueError(f"'{solver_name}' not supported.")

        duration = self.hamiltonian.emulator_ir.duration

        state = np.asarray(state).astype(np.complex128, copy=False)

        solver = ode(self.hamiltonian._ode_real_kernel_int)
        solver.set_f_params(np.zeros_like(state, dtype=np.complex128))
        solver.set_initial_value(state.view(np.float64))
        solver.set_integrator(solver_name, atol=atol, rtol=rtol, nsteps=nsteps)

        if any(time >= duration or time < 0.0 for time in times):
            raise ValueError("Times must be between 0 and duration.")

        times = [*times, duration]

        for time in times:
            if time == 0.0:
                yield state
                continue
            solver.integrate(time)
            AnalogGate._error_check(solver_name, solver.get_return_code())
            u = np.exp(-1j * time * self.hamiltonian.rydberg)
            yield u * solver.y.view(np.complex128)

    @beartype
    def apply(
        self,
        state: StateArray,
        solver_name: str = "dop853",
        atol: float = 1e-7,
        rtol: float = 1e-14,
        nsteps: int = 2_147_483_647,
        times: Union[List[float], RealArray] = [],
        interaction_picture: bool = False,
    ):
        if interaction_picture:
            return self._apply_interation_picture(
                state,
                solver_name=solver_name,
                atol=atol,
                rtol=rtol,
                nsteps=nsteps,
                times=times,
            )
        else:
            return self._apply(
                state,
                solver_name=solver_name,
                atol=atol,
                rtol=rtol,
                nsteps=nsteps,
                times=times,
            )

    @beartype
    def run(
        self,
        shots: int = 1,
        solver_name: str = "dop853",
        atol: float = 1e-14,
        rtol: float = 1e-7,
        nsteps: int = 2_147_483_647,
        interaction_picture: bool = False,
        project_hyperfine: bool = True,
    ):
        """Run the emulation with all atoms in the ground state,
        sampling the final state vector."""

        options = dict(
            solver_name=solver_name,
            atol=atol,
            rtol=rtol,
            nsteps=nsteps,
            interaction_picture=interaction_picture,
        )

        state = self.hamiltonian.space.zero_state()
        (result,) = self.apply(state, **options)
        result /= np.linalg.norm(result)

        return self.hamiltonian.space.sample_state_vector(
            result, shots, project_hyperfine=project_hyperfine
        )



from .hash import HashTrait
from .append import AppendTrait
from .slice import SliceTrait
from .canonicalize import CanonicalizeTrait

__all__ = [
    "HashTrait",
    "AppendTrait",
    "SliceTrait",
    "CanonicalizeTrait",
]



from collections import OrderedDict
from decimal import Decimal
from bloqade.ir.location import ListOfLocations
import tempfile
import pandas as pd
import numpy as np
from bloqade.task.base import Report, Geometry


def test_integration_report():
    prog = (
        ListOfLocations()
        .add_position((0, 0))
        .add_position((0, 6))
        .add_position((0, 3))
        .rydberg.detuning.uniform.piecewise_linear(
            durations=["up_time", "anneal_time", "up_time"],
            values=[
                "initial_detuning",
                "initial_detuning",
                "final_detuning",
                "final_detuning",
            ],
        )
        .rydberg.rabi.amplitude.uniform.piecewise_linear(
            durations=["up_time", "anneal_time", "up_time"],
            values=[0, "rabi_amplitude_max", "rabi_amplitude_max", 0],
        )
        .assign(
            up_time=0.1,
            anneal_time=3.8,
            initial_detuning=-15,
            final_detuning=10,
            rabi_amplitude_max=15,
        )
    )

    with tempfile.NamedTemporaryFile() as f:
        future = prog.quera.mock(state_file=f.name).run_async(shots=100)
        future.pull()
        future2 = future.remove_tasks("Completed")
        future2

    report = future.report()

    bistrings = report.bitstrings()
    counts = report.counts()
    dataframe = report.dataframe
    rydberg_densities = report.rydberg_densities()

    print(bistrings)
    print(counts)
    print(dataframe)
    print(rydberg_densities)


def get_report():
    data = [
        ((0, (0, 0), "11", "11"), np.asarray([1, 1], dtype=np.uint8)),
        ((0, (0, 0), "11", "11"), np.asarray([0, 0], dtype=np.uint8)),
        ((0, (1, 0), "11", "11"), np.asarray([1, 0], dtype=np.uint8)),
        ((0, (1, 0), "11", "11"), np.asarray([1, 0], dtype=np.uint8)),
        ((1, (0, 0), "11", "11"), np.asarray([1, 1], dtype=np.uint8)),
        ((1, (0, 0), "11", "11"), np.asarray([0, 1], dtype=np.uint8)),
        ((1, (1, 0), "11", "10"), np.asarray([1, 0], dtype=np.uint8)),
        ((1, (1, 0), "11", "11"), np.asarray([1, 1], dtype=np.uint8)),
    ]

    metas = [dict(a=1, c=[1, 2]), dict(a=2, b=3, c=[1, 4])]
    geos = [
        Geometry([(0, 0), (0, 6)], [1, 1], None),
        Geometry([(0, 0), (0, 6)], [1, 1], None),
    ]

    index, bitstrings = list(zip(*data))

    index = pd.MultiIndex.from_tuples(
        index, names=["task_number", "cluster", "perfect_sorting", "pre_sequence"]
    )

    dataframe = pd.DataFrame(bitstrings, index=index)

    return Report(dataframe, metas, geos)


def test_list_param():
    report = get_report()

    assert report.list_param("a") == [Decimal("1"), Decimal("2")], "failed, a"
    assert report.list_param("b") == [None, Decimal("3")], "failed, b"
    assert report.list_param("c") == [
        [Decimal("1"), Decimal("2")],
        [Decimal("1"), Decimal("4")],
    ], "failed, c"


def test_bitstrings():
    report = get_report()

    expected_bitstrings = [
        np.array([[1, 1], [0, 0], [1, 0], [1, 0]], dtype=np.uint8),
        np.array([[1, 1], [0, 1], [1, 0], [1, 1]], dtype=np.uint8),
    ]

    bitstrings = report.bitstrings(filter_perfect_filling=False)

    assert all(
        np.equal(lhs, rhs).all() for lhs, rhs in zip(bitstrings, expected_bitstrings)
    ), "failed, no filter"

    expected_bitstrings = [
        np.array([[1, 1], [0, 0], [1, 0], [1, 0]], dtype=np.uint8),
        np.array([[1, 1], [0, 1], [1, 1]], dtype=np.uint8),
    ]

    bitstrings = report.bitstrings(filter_perfect_filling=True)

    assert all(
        np.equal(lhs, rhs).all() for lhs, rhs in zip(bitstrings, expected_bitstrings)
    ), "failed, filter perfect filling"

    expected_bitstrings = [
        np.asarray([[1, 1], [0, 0]], dtype=np.uint8),
        np.asarray([[1, 1], [0, 1]], dtype=np.uint8),
    ]

    bitstrings = report.bitstrings(filter_perfect_filling=True, clusters=(0, 0))

    assert all(
        np.equal(lhs, rhs).all() for lhs, rhs in zip(bitstrings, expected_bitstrings)
    ), "failed, filter perfect filling and cluster (0, 0)"

    expected_bitstrings = [
        np.asarray([[1, 0], [1, 0]], dtype=np.uint8),
        np.asarray([[1, 1]], dtype=np.uint8),
    ]

    bitstrings = report.bitstrings(filter_perfect_filling=True, clusters=(1, 0))

    assert all(
        np.equal(lhs, rhs).all() for lhs, rhs in zip(bitstrings, expected_bitstrings)
    ), "failed, filter perfect filling and cluster (1, 0))"


def test_counts():
    report = get_report()

    counts_0 = OrderedDict()
    counts_1 = OrderedDict()

    counts_0["10"] = 2
    counts_0["00"] = 1
    counts_0["11"] = 1

    counts_1["11"] = 2
    counts_1["01"] = 1
    counts_1["10"] = 1

    expected_counts = [
        counts_0,
        counts_1,
    ]

    counts = report.counts(filter_perfect_filling=False)

    assert counts == expected_counts, "failed, no filter"

    counts_0 = OrderedDict()
    counts_1 = OrderedDict()

    counts_0["10"] = 2
    counts_0["00"] = 1
    counts_0["11"] = 1

    counts_1["11"] = 2
    counts_1["01"] = 1

    expected_counts = [
        counts_0,
        counts_1,
    ]

    counts = report.counts(filter_perfect_filling=True)

    assert counts == expected_counts, "failed, filter perfect filling"

    counts_0 = OrderedDict()
    counts_1 = OrderedDict()

    counts_0["00"] = 1
    counts_0["11"] = 1

    counts_1["01"] = 1
    counts_1["11"] = 1

    expected_counts = [
        counts_0,
        counts_1,
    ]

    counts = report.counts(filter_perfect_filling=True, clusters=(0, 0))

    assert (
        counts == expected_counts
    ), "failed, filter perfect filling and cluster (0, 0)"

    counts_0 = OrderedDict()
    counts_1 = OrderedDict()

    counts_0["10"] = 2

    counts_1["11"] = 1

    expected_counts = [
        counts_0,
        counts_1,
    ]

    counts = report.counts(filter_perfect_filling=True, clusters=(1, 0))

    assert (
        counts == expected_counts
    ), "failed, filter perfect filling and cluster (1, 0)"


def test_rydberg_densities():
    report = get_report()

    index = pd.MultiIndex.from_tuples([(0,), (1,)], names=["task_number"])

    expected_rydberg_densities = [
        np.array([0.25, 0.75]),
        np.array([0.25, 0.25]),
    ]

    expected_rydberg_densities = pd.DataFrame(expected_rydberg_densities, index=index)

    assert np.array_equal(
        expected_rydberg_densities.to_numpy(),
        report.rydberg_densities(filter_perfect_filling=False).to_numpy(),
    ), "failed, no filter"

    expected_rydberg_densities = [
        np.array([0.25, 0.75]),
        np.array([0.3333333, 0.0]),
    ]

    expected_rydberg_densities = pd.DataFrame(expected_rydberg_densities, index=index)

    assert np.allclose(
        expected_rydberg_densities.to_numpy(),
        report.rydberg_densities(filter_perfect_filling=True).to_numpy(),
    ), "failed, filter perfect filling"

    expected_rydberg_densities = [
        np.array([0.5, 0.5]),
        np.array([0.5, 0.0]),
    ]

    expected_rydberg_densities = pd.DataFrame(expected_rydberg_densities, index=index)

    assert np.array_equal(
        expected_rydberg_densities.to_numpy(),
        report.rydberg_densities(
            filter_perfect_filling=True, clusters=(0, 0)
        ).to_numpy(),
    ), "failed, filter perfect filling and cluster (0, 0)"

    expected_rydberg_densities = [
        np.array([0.0, 1.0]),
        np.array([0.0, 0.0]),
    ]

    expected_rydberg_densities = pd.DataFrame(expected_rydberg_densities, index=index)

    assert np.array_equal(
        expected_rydberg_densities.to_numpy(),
        report.rydberg_densities(
            filter_perfect_filling=True, clusters=(1, 0)
        ).to_numpy(),
    ), "failed, filter perfect filling and cluster (1, 0)"



from bloqade.ir.visitor import BloqadeIRVisitor
from bloqade.ir import scalar
from bloqade.ir.control import field, pulse, sequence, waveform
from decimal import Decimal


class ValidatePiecewiseLinearChannel(BloqadeIRVisitor):
    def __init__(
        self,
        level_coupling: sequence.LevelCoupling,
        field_name: pulse.FieldName,
        spatial_modulations: field.SpatialModulation,
    ):
        self.level_coupling = level_coupling
        self.field_name = field_name
        self.spatial_modulations = spatial_modulations

    def check_continuous(
        self,
        time: scalar.Scalar,
        level: str,
        lhs: waveform.Waveform,
        rhs: waveform.Waveform,
    ):
        diff = abs(lhs.eval_decimal(lhs.duration()) - rhs.eval_decimal(Decimal(0)))
        if diff > Decimal("2.2e-16"):
            raise ValueError(
                f"failed to compile waveform to piecewise linear. On the {level} level "
                f"a discontinuity of {diff} was found at time={time()} "
                f"for the {self.level_coupling} {self.field_name} with spatial "
                f"modulation:\n{self.spatial_modulations}"
            )

    def emit_shape_error(self, node: waveform.Waveform):
        return (
            "failed to compile waveform to piecewise linear. "
            f"found non-linear waveform: {node.print_node()} "
            f"for the {self.level_coupling} {self.field_name} with spatial "
            f"modulation:\n{self.spatial_modulations}"
        )

    def visit_waveform_Poly(self, node: waveform.Poly) -> None:
        if len(node.coeffs) > 2:
            raise ValueError(self.emit_shape_error(node))

    def visit_waveform_Sample(self, node: waveform.Sample) -> None:
        if node.interpolation != waveform.Interpolation.Linear:
            raise ValueError(self.emit_shape_error(node))

    def visit_waveform_Smooth(self, node: waveform.Smooth) -> None:
        raise ValueError(self.emit_shape_error(node))

    def visit_waveform_PythonFn(self, node: waveform.PythonFn) -> None:
        raise ValueError(self.emit_shape_error(node))

    def visit_waveform_Append(self, node: waveform.Append) -> None:
        for wf in node.waveforms:
            self.visit(wf)

        result = node.waveforms[0]

        time = result.duration

        for lhs, rhs in zip(node.waveforms[:-1], node.waveforms[1:]):
            self.check_continuous(time, "Waveform", lhs, rhs)
            time = time + lhs.duration

    def visit_sequence_Append(self, node: sequence.Append) -> waveform.Waveform:
        wfs = list(map(self.visit, node.sequences))

        result = wfs[0]

        time = result.duration

        for lhs, rhs in zip(wfs[:-1], wfs[1:]):
            self.check_continuous(time, "Sequence", lhs, rhs)
            result = result.append(rhs)
            time = time + lhs.duration

        return result

    def visit_sequence_Slice(self, node: sequence.Slice) -> waveform.Waveform:
        interval = node.interval
        wf = self.visit(node.sequence)
        return wf[interval.start : interval.stop]

    def visit_sequence_NamedSequence(
        self, node: sequence.NamedSequence
    ) -> waveform.Waveform:
        return self.visit(node.sequence)

    def visit_sequence_Sequence(self, node: sequence.Sequence) -> waveform.Waveform:
        return self.visit(node.pulses[self.level_coupling])

    def visit_pulse_Append(self, node: pulse.Append) -> waveform.Waveform:
        wfs = list(map(self.visit, node.pulses))

        result = wfs[0]

        time = result.duration

        for lhs, rhs in zip(wfs[:-1], wfs[1:]):
            self.check_continuous(time, "Pulse", lhs, rhs)
            result = result.append(rhs)
            time = time + lhs.duration

        return result

    def visit_pulse_Slice(self, node: pulse.Slice) -> waveform.Waveform:
        interval = node.interval
        wf = self.visit(node.pulse)
        return wf[interval.start : interval.stop]

    def visit_pulse_NamedPulse(self, node: pulse.NamedPulse) -> waveform.Waveform:
        return self.visit(node.pulse)

    def visit_pulse_Pulse(self, node: pulse.Pulse) -> waveform.Waveform:
        return self.visit(node.fields[self.field_name])

    def visit_field_Field(self, node: field.Field) -> waveform.Waveform:
        wf = node.drives[self.spatial_modulations]

        self.visit(wf)

        return wf

    # node type doesn't matter here
    def scan(self, node) -> None:
        self.visit(node)



from bloqade.compiler.analysis.common import ScanChannels, ScanVariables, AssignmentScan
from bloqade.compiler.codegen.python.emulator_ir import EmulatorProgramCodeGen
from bloqade.compiler.rewrite.common import (
    AddPadding,
    FlattenCircuit,
    AssignBloqadeIR,
    AssignToLiteral,
    Canonicalizer,
)


def flatten(circuit):
    level_couplings = ScanChannels().scan(circuit)
    circuit = AddPadding(level_couplings=level_couplings).visit(circuit)
    return FlattenCircuit(level_couplings=level_couplings).visit(circuit)


def assign(assignments, circuit):
    completed_assignments = AssignmentScan(assignments).scan(circuit)
    circuit = AssignBloqadeIR(completed_assignments).emit(circuit)
    assignment_analysis = ScanVariables().scan(circuit)

    if not assignment_analysis.is_assigned:
        missing_vars = assignment_analysis.scalar_vars.union(
            assignment_analysis.vector_vars
        )
        raise ValueError(
            "Missing assignments for variables:\n"
            + ("\n".join(f"{var}" for var in missing_vars))
            + "\n"
        )

    return completed_assignments, Canonicalizer().visit(
        AssignToLiteral().visit(circuit)
    )


def generate_emulator_ir(circuit, blockade_radius, waveform_runtime):
    return EmulatorProgramCodeGen(
        blockade_radius=blockade_radius, waveform_runtime=waveform_runtime
    ).emit(circuit)



from pydantic import BaseModel
from decimal import Decimal
from beartype.typing import Optional
from beartype import beartype

__all__ = ["QuEraCapabilities"]


class RydbergGlobalCapabilities(BaseModel):
    rabi_frequency_min: Decimal
    rabi_frequency_max: Decimal
    rabi_frequency_resolution: Decimal
    rabi_frequency_slew_rate_max: Decimal
    detuning_min: Decimal
    detuning_max: Decimal
    detuning_resolution: Decimal
    detuning_slew_rate_max: Decimal
    phase_min: Decimal
    phase_max: Decimal
    phase_resolution: Decimal
    time_min: Decimal
    time_max: Decimal
    time_resolution: Decimal
    time_delta_min: Decimal

    @beartype
    def scale_units(
        self, _: Decimal, energy_scale: Decimal
    ) -> "RydbergGlobalCapabilities":
        return RydbergGlobalCapabilities(
            rabi_frequency_min=self.rabi_frequency_min * energy_scale,
            rabi_frequency_max=self.rabi_frequency_max * energy_scale,
            rabi_frequency_resolution=self.rabi_frequency_resolution * energy_scale,
            rabi_frequency_slew_rate_max=self.rabi_frequency_slew_rate_max
            * energy_scale**2,
            detuning_min=self.detuning_min * energy_scale,
            detuning_max=self.detuning_max * energy_scale,
            detuning_resolution=self.detuning_resolution * energy_scale,
            detuning_slew_rate_max=self.detuning_slew_rate_max * energy_scale**2,
            phase_min=self.phase_min,
            phase_max=self.phase_max,
            phase_resolution=self.phase_resolution,
            time_min=self.time_min / energy_scale,
            time_max=self.time_max / energy_scale,
            time_resolution=self.time_resolution / energy_scale,
            time_delta_min=self.time_delta_min / energy_scale,
        )


class RydbergLocalCapabilities(BaseModel):
    detuning_min: Decimal
    detuning_max: Decimal
    detuning_slew_rate_max: Decimal
    site_coefficient_min: Decimal
    site_coefficient_max: Decimal
    number_local_detuning_sites: int
    spacing_radial_min: Decimal
    time_resolution: Decimal
    time_delta_min: Decimal

    @beartype
    def scale_units(self, distance_scale, energy_scale) -> "RydbergLocalCapabilities":
        return RydbergLocalCapabilities(
            detuning_min=self.detuning_min * energy_scale,
            detuning_max=self.detuning_max * energy_scale,
            detuning_slew_rate_max=self.detuning_slew_rate_max * energy_scale**2,
            site_coefficient_min=self.site_coefficient_min,
            site_coefficient_max=self.site_coefficient_max,
            number_local_detuning_sites=self.number_local_detuning_sites,
            spacing_radial_min=self.spacing_radial_min * distance_scale,
            time_resolution=self.time_resolution / energy_scale,
            time_delta_min=self.time_delta_min / energy_scale,
        )


class RydbergCapabilities(BaseModel):
    c6_coefficient: Decimal
    global_: RydbergGlobalCapabilities
    local: Optional[RydbergLocalCapabilities]

    class Config:
        allow_population_by_field_name = True
        fields = {"global_": "global"}

    @beartype
    def scale_units(self, distance_scale, energy_scale) -> "RydbergCapabilities":
        return RydbergCapabilities(
            c6_coefficient=self.c6_coefficient * energy_scale * distance_scale**6,
            global_=self.global_.scale_units(distance_scale, energy_scale),
            local=self.local.scale_units(distance_scale, energy_scale),
        )


class LatticeGeometryCapabilities(BaseModel):
    spacing_radial_min: Decimal
    spacing_vertical_min: Decimal
    position_resolution: Decimal
    number_sites_max: int

    @beartype
    def scale_units(
        self, distance_scale: Decimal, _: Decimal
    ) -> "LatticeGeometryCapabilities":
        return LatticeGeometryCapabilities(
            spacing_radial_min=self.spacing_radial_min * distance_scale,
            spacing_vertical_min=self.spacing_vertical_min * distance_scale,
            position_resolution=self.position_resolution * distance_scale,
            number_sites_max=self.number_sites_max,
        )


class LatticeAreaCapabilities(BaseModel):
    width: Decimal
    height: Decimal

    @beartype
    def scale_units(
        self, distance_scale: Decimal, _: Decimal
    ) -> "LatticeAreaCapabilities":
        return LatticeAreaCapabilities(
            width=self.width * distance_scale,
            height=self.height * distance_scale,
        )


class LatticeCapabilities(BaseModel):
    number_qubits_max: int
    area: LatticeAreaCapabilities
    geometry: LatticeGeometryCapabilities

    def scale_units(
        self, distance_scale: Decimal, energy_scale: Decimal
    ) -> "LatticeCapabilities":
        return LatticeCapabilities(
            number_qubits_max=self.number_qubits_max,
            area=self.area.scale_units(distance_scale, energy_scale),
            geometry=self.geometry.scale_units(distance_scale, energy_scale),
        )


class TaskCapabilities(BaseModel):
    number_shots_min: int
    number_shots_max: int

    @beartype
    def scale_units(self, _: Decimal, __: Decimal) -> "TaskCapabilities":
        return TaskCapabilities(
            number_shots_min=self.number_shots_min,
            number_shots_max=self.number_shots_max,
        )


class DeviceCapabilities(BaseModel):
    task: TaskCapabilities
    lattice: LatticeCapabilities
    rydberg: RydbergCapabilities

    @beartype
    def scale_units(
        self, distance_scale: Decimal, energy_scale: Decimal
    ) -> "DeviceCapabilities":
        return DeviceCapabilities(
            task=self.task.scale_units(distance_scale, energy_scale),
            lattice=self.lattice.scale_units(distance_scale, energy_scale),
            rydberg=self.rydberg.scale_units(distance_scale, energy_scale),
        )


class QuEraCapabilities(BaseModel):
    version: str
    capabilities: DeviceCapabilities

    @beartype
    def scale_units(
        self, distance_scale: Decimal, energy_scale: Decimal
    ) -> "QuEraCapabilities":
        return QuEraCapabilities(
            version=self.version,
            capabilities=self.capabilities.scale_units(distance_scale, energy_scale),
        )



from functools import cached_property
from bloqade.ir.scalar import Scalar, cast


class SliceTrait:
    @property
    def _sub_expr(self):
        raise NotImplementedError("sub_expr property is not implemented")

    @cached_property
    def start(self) -> Scalar:
        """Start time of the sliced object

        Returns:
            Scalar: The starting time of the sliced object as a
            Scalar Expression
        """
        if self.interval.start is None:
            return cast(0)
        else:
            return self.interval.start

    @cached_property
    def stop(self) -> Scalar:
        """Stop time of the sliced object

        Returns:
            Scalar: The stopping time of the sliced object as a
            Scalar Expression
        """
        if self.interval.stop is None:
            return self._sub_expr.duration
        else:
            return self.interval.stop

    @cached_property
    def duration(self) -> Scalar:
        return self._sub_expr.duration[self.interval.start : self.interval.stop]



from bloqade import start
from bloqade.submission.capabilities import get_capabilities
from bloqade.ir import analog_circuit

from bloqade.compiler.passes.hardware import (
    analyze_channels,
    canonicalize_circuit,
    assign_circuit,
    validate_waveforms,
    generate_ahs_code,
)


circuit = (
    start.rydberg.detuning.uniform.piecewise_linear([0.1, 1.2, 0.3], [-10, -10, 10, 10])
    .amplitude.uniform.piecewise_linear([0.1, 1.4, 0.1], [0, 10, 10, 0])
    .parse_circuit()
)

seq2 = start.rydberg.detuning.uniform.piecewise_linear(
    [0.3, 1.2, 0.3], [10, 10, -10, -10]
).parse_sequence()

circuit = analog_circuit.AnalogCircuit(circuit.register, circuit.sequence.append(seq2))

assignments = {}
capabilities = get_capabilities()


level_couplings = analyze_channels(circuit)
circuit = canonicalize_circuit(circuit, level_couplings)
circuit = assign_circuit(circuit, assignments)
validate_waveforms(level_couplings, circuit)
ahs_code = generate_ahs_code(capabilities, level_couplings, circuit)
# 7. specialize to QuEra/Braket IR
quera_ir, parallel_decoder = ahs_code.generate_quera_ir(shots=100)
braket_ir, parallel_decoder = ahs_code.generate_braket_ir(shots=100)
print(quera_ir.effective_hamiltonian.rydberg.rabi_frequency_amplitude.global_)



from functools import cached_property
from bloqade.ir.scalar import cast


class AppendTrait:
    @property
    def _sub_exprs(self):
        raise NotImplementedError("sub_exprs property is not implemented")

    @cached_property
    def duration(self):
        duration = cast(0)
        for p in self._sub_exprs:
            duration = duration + p.duration

        return duration



import bloqade.ir.control.field as field
import bloqade.ir.control.waveform as waveform
import bloqade.ir.scalar as scalar
from bloqade.builder.typing import LiteralType
from bloqade.ir.visitor import BloqadeIRTransformer
from typing import Any, Dict


class AssignBloqadeIR(BloqadeIRTransformer):
    def __init__(self, mapping: Dict[str, LiteralType]):
        self.mapping = dict(mapping)

    def visit_scalar_Variable(self, node: scalar.Variable):
        if node.name in self.mapping:
            return scalar.AssignedVariable(node.name, self.mapping[node.name])

        return node

    def visit_scalar_AssignedVariable(self, node: scalar.AssignedVariable):
        if node.name in self.mapping:
            raise ValueError(f"Variable {node.name} already assigned to {node.value}.")

        return node

    def visit_field_RunTimeVector(self, node: field.RunTimeVector):
        if node.name in self.mapping:
            return field.AssignedRunTimeVector(node.name, self.mapping[node.name])

        return node

    def visit_waveform_Record(self, node: waveform.Record):
        if node.var.name in self.mapping:
            return self.visit(node.waveform)

        return waveform.Record(self.visit(node.waveform), node.var)

    def visit_field_AssignedRunTimeVector(self, node: field.AssignedRunTimeVector):
        if node.name in self.mapping:
            raise ValueError(f"Variable {node.name} already assigned to {node.value}.")

        return node

    def emit(self, node) -> Any:
        return self.visit(node)



import bloqade.ir.control.waveform as waveform
import bloqade.ir.control.field as field
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.sequence as sequence
from bloqade.builder.typing import LiteralType
from bloqade.ir.visitor import BloqadeIRVisitor

from decimal import Decimal
from beartype.typing import Any, Dict
from beartype import beartype


class IsConstant(BloqadeIRVisitor):
    @beartype
    def __init__(self, assignments: Dict[str, LiteralType] = {}) -> None:
        self.assignments = dict(assignments)
        self.is_constant = True

    def generic_visit(self, node: Any) -> Any:
        # skip visiting children if we already know it's not constant
        if not self.is_constant:
            return

        super().generic_visit(node)

    def visit_waveform_Linear(self, node: waveform.Linear):
        diff = node.stop(**self.assignments) - node.start(**self.assignments)
        self.is_constant = self.is_constant and (diff == 0)

    def visit_waveform_Poly(self, node: waveform.Poly):
        coeffs = [coeff(**self.assignments) for coeff in node.coeffs]
        if any(coeff != 0 for coeff in coeffs[1:]):
            self.is_constant = False

    def visit_waveform_PythonFn(self, node: waveform.PythonFn):
        # can't analyze python functions, assume it's not constant
        self.is_constant = False

    def visit_waveform_Append(self, node: waveform.Append):
        value = None
        for wf in node.waveforms:
            is_constant = IsConstant(self.assignments).scan(wf)

            if not is_constant:
                self.is_constant = False
                return

            if value is None:
                value = wf.eval_decimal(Decimal("0"), **self.assignments)
                new_value = value
            else:
                new_value = wf.eval_decimal(Decimal("0"), **self.assignments)

            if new_value != value:
                self.is_constant = False
                return

    def visit_waveform_Add(self, node: waveform.Add):
        left_duration = node.left.duration(**self.assignments)
        right_duration = node.right.duration(**self.assignments)

        if left_duration != right_duration:
            self.is_constant = False
            return

        self.visit(node.left)
        self.visit(node.right)

    def visit_pulse_Pulse(self, node: pulse.Pulse):
        duration = node.duration(**self.assignments)
        for val in node.fields.values():
            self.visit(val)

            if not self.is_constant:
                return

            if val.duration(**self.assignments) != duration:
                self.is_constant = False
                return

    def visit_pulse_Field(self, node: field.Field):
        duration = node.duration(**self.assignments)
        for val in node.drives.values():
            self.visit(val)

            if not self.is_constant:
                return

            if val.duration(**self.assignments) != duration:
                self.is_constant = False
                return

    def visit_sequence_Sequence(self, node: sequence.Sequence):
        duration = node.duration(**self.assignments)
        for val in node.pulses.values():
            self.visit(val)

            if not self.is_constant:
                return

            if val.duration(**self.assignments) != duration:
                self.is_constant = False
                return

    def scan(self, node) -> bool:
        self.visit(node)
        return self.is_constant



from .assignment_scan import AssignmentScan
from .check_slices import CheckSlices
from .is_constant import IsConstant
from .is_hyperfine import IsHyperfineSequence
from .scan_channels import ScanChannels
from .scan_variables import ScanVariables


__all__ = [
    "AssignmentScan",
    "CheckSlices",
    "IsConstant",
    "IsHyperfineSequence",
    "ScanChannels",
    "ScanVariables",
]



import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.field as field
import bloqade.ir.control.pulse as pulse

from bloqade.ir.visitor import BloqadeIRVisitor


class ScanChannels(BloqadeIRVisitor):
    def __init__(self):
        self.channels = None

    def visit_sequence_Sequence(self, node: sequence.Sequence):
        for lc, p in node.pulses.items():
            saved = dict() if self.channels is None else dict(self.channels)
            self.channels = saved.get(lc, {})
            self.visit(p)
            saved[lc] = self.channels
            self.channels = dict(saved)

    def visit_pulse_Pulse(self, node: pulse.Pulse):
        for fn, f in node.fields.items():
            saved = dict() if self.channels is None else dict(self.channels)
            self.channels = saved.get(fn, set())
            self.visit(f)
            saved[fn] = self.channels
            self.channels = dict(saved)

    def visit_field_Field(self, node: field.Field):
        self.channels = set() if self.channels is None else self.channels
        self.channels = set(node.drives.keys()) | self.channels

    def scan(self, node):
        self.visit(node)
        return self.channels



from decimal import Decimal
import pytest
from bloqade.compiler.analysis.hardware.piecewise_linear import (
    ValidatePiecewiseLinearChannel,
)
from bloqade.compiler.analysis.hardware.piecewise_constant import (
    ValidatePiecewiseConstantChannel,
)
from bloqade.compiler.analysis.hardware.channels import ValidateChannels
from bloqade.ir import analog_circuit
from bloqade.compiler.rewrite.common.add_padding import AddPadding

import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.field as field
import bloqade.ir.control.waveform as waveform

from bloqade import piecewise_constant, piecewise_linear, var, start


@waveform.to_waveform(1)
def py_func(x):
    return x


def test_validate_channels_happy_path():
    wf1 = waveform.Constant(1, 1)
    wf2 = waveform.Linear(1, 1, 1)

    sm = field.AssignedRunTimeVector("mask", [Decimal("1.0"), Decimal("2.0")])

    f1 = field.Field({field.Uniform: wf1, sm: wf2})
    f2 = field.Field({sm: wf2})
    f3 = f1.add(f2)

    p1 = pulse.Pulse({pulse.detuning: f1})
    p2 = pulse.Pulse({pulse.detuning: f2})
    p3 = pulse.Pulse({pulse.detuning: f3})

    s1 = sequence.Sequence({sequence.rydberg: p1})
    s2 = sequence.Sequence({sequence.rydberg: p2})
    s3 = sequence.Sequence({sequence.rydberg: p3})

    register = start.add_position([(0, 0), (1, 1)])
    circuit1 = analog_circuit.AnalogCircuit(register, s1)
    circuit2 = analog_circuit.AnalogCircuit(register, s2)
    circuit3 = analog_circuit.AnalogCircuit(register, s3)

    assert ValidateChannels().scan(circuit1) is None
    assert ValidateChannels().scan(circuit2) is None
    assert ValidateChannels().scan(circuit3) is None


def test_validate_channels_sad_path():
    wf1 = waveform.Constant(1, 1)
    wf2 = waveform.Linear(1, 1, 1)
    wf3 = waveform.Poly([1, 2], 1)

    sm1 = field.AssignedRunTimeVector(
        "mask", [Decimal("1.0"), Decimal("2.0"), Decimal("3.0")]
    )
    sm2 = field.ScaledLocations.create({0: Decimal("1.0"), 1: Decimal("2.0")})
    sm2_broken = field.ScaledLocations.create({1: Decimal("1.0"), 3: Decimal("2.0")})
    sm1_broken = field.AssignedRunTimeVector("mask", [Decimal("1.9"), Decimal("1.0")])

    f0 = field.Field({field.Uniform: wf1, sm1: wf2})
    f1 = field.Field({field.Uniform: wf1, sm1: wf2, sm2: wf3})
    f2 = field.Field({field.Uniform: wf1, sm2_broken: wf2})
    f3 = field.Field({field.Uniform: wf1, sm1_broken: wf2})

    p0 = pulse.Pulse({pulse.detuning: f0})
    p1 = pulse.Pulse({pulse.detuning: f1})
    p2 = pulse.Pulse({pulse.detuning: f2})
    p3 = pulse.Pulse({pulse.detuning: f3})
    p4 = pulse.Pulse({pulse.rabi.amplitude: f1, pulse.detuning: f0})

    s1 = sequence.Sequence({sequence.rydberg: p1})
    s2 = sequence.Sequence({sequence.rydberg: p2})
    s3 = sequence.Sequence({sequence.rydberg: p3})
    s4 = sequence.Sequence({sequence.hyperfine: p0, sequence.rydberg: p0})
    s5 = sequence.Sequence({sequence.rydberg: p4})

    register = start.add_position([(0, 0), (1, 1), (2, 2)])
    circuit1 = analog_circuit.AnalogCircuit(register, s1)
    circuit2 = analog_circuit.AnalogCircuit(register, s2)
    circuit3 = analog_circuit.AnalogCircuit(register, s3)
    circuit4 = analog_circuit.AnalogCircuit(register, s4)
    circuit5 = analog_circuit.AnalogCircuit(register, s5)

    with pytest.raises(ValueError):
        ValidateChannels().scan(circuit1)

    with pytest.raises(ValueError):
        ValidateChannels().scan(circuit2)

    with pytest.raises(ValueError):
        ValidateChannels().scan(circuit3)

    with pytest.raises(ValueError):
        ValidateChannels().scan(circuit4)

    with pytest.raises(ValueError):
        ValidateChannels().scan(circuit5)


def test_piecewise_constant_waveform_happy_path():
    validator = ValidatePiecewiseConstantChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )

    wf1 = waveform.Constant(1, 1)
    wf2 = waveform.Linear(1, 1, 1)
    wf3 = waveform.Poly([], 1)
    wf4 = waveform.Poly([1], 1)
    wf5 = py_func.sample(0.1, "constant")

    assert validator.scan(wf1) is None
    assert validator.scan(wf2) is None
    assert validator.scan(wf3) is None
    assert validator.scan(wf4) is None
    assert validator.scan(wf5) is None
    assert validator.scan(wf1.append(wf2).append(wf3)) is None


def test_piecewise_constant_waveform_sad_path():
    validator = ValidatePiecewiseConstantChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )

    wf0 = piecewise_constant([1, 2, 3], [1, 2, 3])

    wf1 = waveform.Linear(1, 2, 1)
    wf2 = waveform.Poly([1, 2, 3], 1)
    wf3 = py_func.sample(0.1, "linear")
    wf4 = py_func
    wf5 = wf2.smooth(1, "Gaussian")
    wf6 = wf0.append(wf1)

    with pytest.raises(ValueError):
        validator.scan(wf1)

    with pytest.raises(ValueError):
        validator.scan(wf2)

    with pytest.raises(ValueError):
        validator.scan(wf3)

    with pytest.raises(ValueError):
        validator.scan(wf4)

    with pytest.raises(ValueError):
        validator.scan(wf5)

    with pytest.raises(ValueError):
        validator.scan(wf6)


def test_piecewise_linear_waveform_happy_path():
    validator = ValidatePiecewiseLinearChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )

    wf1 = waveform.Constant(1, 1)
    wf2 = waveform.Linear(1, 2, 1)
    wf3 = waveform.Poly([], 1)
    wf4 = waveform.Poly([1], 1)
    wf5 = waveform.Poly([1, 2], 1)
    wf6 = py_func.sample(0.1, "linear")
    wf7 = piecewise_linear([0.1, 0.2, 0.3], [1, 2, 3, 4])
    wf8 = wf7[0.05:0.25]
    wf9 = wf1 + wf2

    assert validator.scan(wf1) is None
    assert validator.scan(wf2) is None
    assert validator.scan(wf3) is None
    assert validator.scan(wf4) is None
    assert validator.scan(wf5) is None
    assert validator.scan(wf6) is None
    assert validator.scan(wf7) is None
    assert validator.scan(wf8) is None
    assert validator.scan(wf9) is None


def test_piecewise_linear_waveform_sad_path():
    validator = ValidatePiecewiseLinearChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )

    wf1 = py_func
    wf2 = py_func.sample(0.1, "constant")
    wf3 = waveform.Poly([1, 2, 3], 1)
    wf4 = py_func.smooth(1, "Gaussian")
    wf5 = AddPadding().visit(waveform.Linear(1, 2, 0.5) + waveform.Linear(2, 3, 1))
    wf6 = waveform.Linear(1, 2, 0.5).append(waveform.Linear(2.1, 3, 1))

    with pytest.raises(ValueError):
        validator.scan(wf1)

    with pytest.raises(ValueError):
        validator.scan(wf2)

    with pytest.raises(ValueError):
        validator.scan(wf3)

    with pytest.raises(ValueError):
        validator.scan(wf4)

    with pytest.raises(ValueError):
        validator.scan(wf5)

    with pytest.raises(ValueError):
        validator.scan(wf6)


def test_pulse_happy_path():
    linear = ValidatePiecewiseLinearChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )
    constant = ValidatePiecewiseConstantChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )

    p1 = pulse.Pulse(
        {pulse.rabi.amplitude: field.Field({field.Uniform: waveform.Constant(1, 1)})}
    )
    p2 = pulse.Pulse(
        {pulse.rabi.amplitude: field.Field({field.Uniform: waveform.Linear(1, 0, 1)})}
    )
    p3 = pulse.Pulse(
        {
            pulse.rabi.amplitude: field.Field(
                {field.Uniform: py_func.sample(0.1, "constant")}
            )
        }
    )

    p_test = pulse.NamedPulse("test", (p1.append(p2))[: var("t")])
    assert linear.scan(p_test) is None
    assert constant.scan(p3) is None


def test_pulse_sad_path():
    linear = ValidatePiecewiseLinearChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )
    constant = ValidatePiecewiseConstantChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )

    p1 = pulse.Pulse(
        {pulse.rabi.amplitude: field.Field({field.Uniform: waveform.Constant(1, 1)})}
    )
    p2 = pulse.Pulse(
        {pulse.rabi.amplitude: field.Field({field.Uniform: waveform.Linear(1, 0, 1)})}
    )
    p3 = pulse.Pulse({pulse.rabi.amplitude: field.Field({field.Uniform: py_func})})
    p4 = pulse.Pulse(
        {
            pulse.rabi.amplitude: field.Field(
                {field.Uniform: py_func.sample(0.1, "linear")}
            )
        }
    )

    p1_test = p1.append(p2).append(p3)

    with pytest.raises(ValueError):
        linear.scan(p1_test)

    with pytest.raises(ValueError):
        constant.scan(p4)


def test_sequence_happy_path():
    linear = ValidatePiecewiseLinearChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )
    constant = ValidatePiecewiseConstantChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )

    s1 = sequence.Sequence(
        {
            sequence.rydberg: pulse.Pulse(
                {
                    pulse.rabi.amplitude: field.Field(
                        {field.Uniform: waveform.Constant(1, 1)}
                    )
                }
            )
        }
    )
    s2 = sequence.Sequence(
        {
            sequence.rydberg: pulse.Pulse(
                {
                    pulse.rabi.amplitude: field.Field(
                        {field.Uniform: waveform.Linear(1, 0, 1)}
                    )
                }
            )
        }
    )
    s3 = sequence.Sequence(
        {
            sequence.rydberg: pulse.Pulse(
                {
                    pulse.rabi.amplitude: field.Field(
                        {field.Uniform: py_func.sample(0.1, "constant")}
                    )
                }
            )
        }
    )

    p_test = sequence.NamedSequence(name="test", sequence=(s1.append(s2))[: var("t")])
    assert linear.scan(p_test) is None
    assert constant.scan(s3) is None


def test_sequence_sad_path():
    linear = ValidatePiecewiseLinearChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )
    constant = ValidatePiecewiseConstantChannel(
        level_coupling=sequence.rydberg,
        field_name=pulse.rabi.amplitude,
        spatial_modulations=field.Uniform,
    )

    p1 = sequence.Sequence(
        {
            sequence.rydberg: pulse.Pulse(
                {
                    pulse.rabi.amplitude: field.Field(
                        {field.Uniform: waveform.Constant(1, 1)}
                    )
                }
            )
        }
    )
    p2 = sequence.Sequence(
        {
            sequence.rydberg: pulse.Pulse(
                {
                    pulse.rabi.amplitude: field.Field(
                        {field.Uniform: waveform.Linear(1, 0, 1)}
                    )
                }
            )
        }
    )
    p3 = sequence.Sequence(
        {
            sequence.rydberg: pulse.Pulse(
                {pulse.rabi.amplitude: field.Field({field.Uniform: py_func})}
            )
        }
    )
    p4 = sequence.Sequence(
        {
            sequence.rydberg: pulse.Pulse(
                {
                    pulse.rabi.amplitude: field.Field(
                        {field.Uniform: py_func.sample(0.1, "linear")}
                    )
                }
            )
        }
    )

    p1_test = p1.append(p2).append(p3)

    with pytest.raises(ValueError):
        linear.scan(p1_test)

    with pytest.raises(ValueError):
        constant.scan(p4)



from collections import OrderedDict
from functools import cached_property

from bloqade.ir.scalar import Interval, Scalar, cast
from bloqade.ir.tree_print import Printer
from bloqade.ir.control.field import Field
from bloqade.ir.control.traits import (
    HashTrait,
    AppendTrait,
    SliceTrait,
    CanonicalizeTrait,
)
from beartype.typing import List
from pydantic.dataclasses import dataclass
from bloqade.visualization import get_pulse_figure
from bloqade.visualization import display_ir


__all__ = [
    "Pulse",
    "NamedPulse",
    "FieldName",
    "rabi",
    "detuning",
]


@dataclass(frozen=True)
class FieldName:
    def print_node(self):
        return self.__class__.__name__

    def children(self):
        return []

    def __str__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def __eq__(self, other):
        return self.__class__ == other.__class__


@dataclass(frozen=True)
class RabiFrequencyAmplitude(FieldName):
    def print_node(self):
        return "RabiFrequencyAmplitude"


@dataclass(frozen=True)
class RabiFrequencyPhase(FieldName):
    def print_node(self):
        return "RabiFrequencyPhase"


@dataclass(frozen=True)
class Detuning(FieldName):
    def print_node(self):
        return "Detuning"


class RabiRouter:
    def __init__(self) -> None:
        self.amplitude = RabiFrequencyAmplitude()
        self.phase = RabiFrequencyPhase()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def print_node(self):
        return "RabiRouter"

    def children(self):
        return {"Amplitude": self.amplitude, "Phase": self.phase}


rabi = RabiRouter()
detuning = Detuning()


@dataclass(frozen=True)
class PulseExpr(HashTrait, CanonicalizeTrait):
    """
    ```bnf
    <expr> ::= <pulse>
      | <append>
      | <slice>
      | <named>
    ```
    """

    __hash__ = HashTrait.__hash__

    def append(self, other: "PulseExpr") -> "PulseExpr":
        return PulseExpr.canonicalize(Append([self, other]))

    def __getitem__(self, s: slice) -> "PulseExpr":
        return PulseExpr.canonicalize(Slice(self, Interval.from_slice(s)))

    def __str__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def _get_data(self, **assigments):
        return NotImplementedError

    def figure(self, **assignments):
        return NotImplementedError

    def show(self, **assignments):
        return NotImplementedError


@dataclass(frozen=True)
class Append(AppendTrait, PulseExpr):
    """
    ```bnf
    <append> ::= <expr>+
    ```
    """

    pulses: List[PulseExpr]

    __hash__ = PulseExpr.__hash__

    @property
    def _sub_exprs(self):
        return self.pulses

    def print_node(self):
        return "Append"

    def children(self):
        return self.pulses


@dataclass(frozen=True)
class Pulse(PulseExpr):
    """
    ```bnf
    <pulse> ::= (<field name> <field>)+
    ```
    """

    fields: dict[FieldName, Field]

    __hash__ = PulseExpr.__hash__

    @staticmethod
    def create(fields) -> "Pulse":
        processed_fields = dict()
        for k, v in fields.items():
            if isinstance(v, Field):
                processed_fields[k] = v
            elif isinstance(v, dict):
                processed_fields[k] = Field(v)
            else:
                raise TypeError(f"Expected Field or dict, got {type(v)}")

        return Pulse(processed_fields)

    @cached_property
    def duration(self) -> Scalar:
        if len(self.fields) == 0:
            return cast(0)
        # Fields are all aligned so that they all start at 0.
        fields = list(self.fields.values())
        duration = fields[0].duration
        for val in fields[1:]:
            duration = duration.max(val.duration)

        return duration

    def print_node(self):
        return "Pulse"

    def children(self):
        # annotated children
        return {k.print_node(): v for k, v in self.fields.items()}

    def _get_data(self, **assigments):
        return None, self.fields

    def figure(self, **assignments):
        return get_pulse_figure(self, **assignments)

    def show(self, **assignments):
        """
        Interactive visualization of the Pulse

        Args:
            **assignments: assigning the instance value (literal) to the
                existing variables in the Pulse

        """
        display_ir(self, assignments)


@dataclass(frozen=True)
class NamedPulse(PulseExpr):
    name: str
    pulse: PulseExpr

    __hash__ = PulseExpr.__hash__

    @cached_property
    def duration(self) -> Scalar:
        return self.pulse.duration

    def print_node(self):
        return "NamedPulse"

    def children(self):
        return OrderedDict([("name", self.name), ("pulse", self.pulse)])

    def _get_data(self, **assigments):
        return self.name, self.pulse.value

    def figure(self, **assignments):
        return get_pulse_figure(self, **assignments)

    def show(self, **assignments):
        display_ir(self, assignments)


@dataclass(frozen=True)
class Slice(SliceTrait, PulseExpr):
    pulse: PulseExpr
    interval: Interval

    __hash__ = PulseExpr.__hash__

    @property
    def _sub_expr(self):
        return self.pulse

    def print_node(self):
        return "Slice"

    def children(self):
        return [self.interval, self.pulse]



from pydantic.dataclasses import dataclass
import bloqade.ir.scalar as scalar
import bloqade.ir.control.field as field

from beartype.typing import FrozenSet

from bloqade.ir.visitor import BloqadeIRVisitor


@dataclass(frozen=True)
class ScanVariableResults:
    scalar_vars: FrozenSet[str]
    vector_vars: FrozenSet[str]
    assigned_scalar_vars: FrozenSet[str]
    assigned_vector_vars: FrozenSet[str]

    @property
    def is_assigned(self) -> bool:
        return len(self.scalar_vars) == 0 or len(self.vector_vars) == 0


class ScanVariables(BloqadeIRVisitor):
    def __init__(self):
        self.scalar_vars = set()
        self.assign_scalar_vars = set()
        self.vector_vars = set()
        self.assigned_vector_vars = set()

    def visit_scalar_Variable(self, node: scalar.Variable):
        self.scalar_vars.add(node.name)

    def visit_scalar_AssignedVariable(self, node: scalar.AssignedVariable):
        self.assign_scalar_vars.add(node.name)

    def visit_field_RunTimeVector(self, node: field.RunTimeVector):
        self.vector_vars.add(node.name)

    def visit_field_AssignedRunTimeVector(self, node: field.AssignedRunTimeVector):
        if node.name is not None:  # skip literal vectors
            self.assigned_vector_vars.add(node.name)

    def scan(self, node) -> ScanVariableResults:
        self.visit(node)
        return ScanVariableResults(
            self.scalar_vars,
            self.vector_vars,
            self.assign_scalar_vars,
            self.assigned_vector_vars,
        )



from collections import OrderedDict
from functools import cached_property
from bloqade.ir.control.pulse import PulseExpr, Pulse
from bloqade.ir.control.traits import (
    HashTrait,
    AppendTrait,
    SliceTrait,
    CanonicalizeTrait,
)
from bloqade.ir.scalar import Interval, Scalar, cast
from bloqade.ir.tree_print import Printer

from pydantic.dataclasses import dataclass
from beartype.typing import List, Dict
from bloqade.visualization import get_ir_figure
from bloqade.visualization import display_ir


__all__ = [
    "LevelCoupling",
    "rydberg",
    "hyperfine",
    "Sequence",
    "NamedSequence",
]


@dataclass(frozen=True)
class LevelCoupling:
    def children(self):
        return []

    def __str__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def __eq__(self, other):
        return self.__class__ == other.__class__


@dataclass(frozen=True)
class RydbergLevelCoupling(LevelCoupling):
    def print_node(self):
        return "RydbergLevelCoupling"

    def __str__(self):
        return "Rydberg"


@dataclass(frozen=True)
class HyperfineLevelCoupling(LevelCoupling):
    def print_node(self):
        return "HyperfineLevelCoupling"

    def __str__(self):
        return "Hyperfine"


rydberg = RydbergLevelCoupling()
hyperfine = HyperfineLevelCoupling()


@dataclass(frozen=True)
class SequenceExpr(HashTrait, CanonicalizeTrait):
    __hash__ = HashTrait.__hash__

    def append(self, other: "SequenceExpr") -> "SequenceExpr":
        return SequenceExpr.canonicalize(Append([self, other]))

    def set_name(self, name: str):  # no need to call canonicalize here
        return NamedSequence(self, name)

    def __getitem__(self, s: slice) -> "Slice":
        return self.canonicalize(Slice(self, Interval.from_slice(s)))

    def __str__(self) -> str:
        ph = Printer()
        ph.print(self)
        return ph.get_value()

    def _repr_pretty_(self, p, cycle):
        Printer(p).print(self, cycle)

    def _get_data(self, **assignment):
        raise NotImplementedError

    def figure(self, **assignment):
        raise NotImplementedError

    def show(self, **assignment):
        raise NotImplementedError


@dataclass(frozen=True)
class Append(AppendTrait, SequenceExpr):
    sequences: List[SequenceExpr]

    __hash__ = SequenceExpr.__hash__

    @property
    def _sub_exprs(self):
        return self.sequences

    def children(self):
        return self.sequences

    def print_node(self):
        return "Append"


@dataclass(frozen=True)
class Sequence(SequenceExpr):
    """Sequence of a program, which includes pulses informations."""

    pulses: dict[LevelCoupling, PulseExpr]

    __hash__ = SequenceExpr.__hash__

    @staticmethod
    def create(pulses: Dict = {}) -> "Sequence":
        processed_pulses = {}
        for level_coupling, pulse in pulses.items():
            if not isinstance(level_coupling, LevelCoupling):
                raise TypeError(f"Unexpected type {type(level_coupling)}")

            if isinstance(pulse, PulseExpr):
                processed_pulses[level_coupling] = pulse
            elif isinstance(pulse, dict):
                processed_pulses[level_coupling] = Pulse.create(pulse)
            else:
                raise TypeError(f"Unexpected type {type(pulse)}")
        return Sequence(processed_pulses)

    @cached_property
    def duration(self) -> Scalar:
        # Pulses are all aligned so that they all start at 0.
        if len(self.pulses) == 0:
            return cast(0)

        pulses = list(self.pulses.values())
        duration = pulses[0].duration
        for p in pulses[1:]:
            duration = duration.max(p.duration)

        return duration

    def __call__(self, clock_s: float, level_coupling: LevelCoupling, *args, **kwargs):
        return self.pulses[level_coupling](clock_s, *args, **kwargs)

    # return annotated version
    def children(self):
        return {k.print_node(): v for k, v in self.pulses.items()}

    def print_node(self):
        return "Sequence"

    def _get_data(self, **assignments):
        return None, self.pulses

    def figure(self, **assignments):
        return get_ir_figure(self, **assignments)

    def show(self, **assignments):
        """
        Interactive visualization of the Sequence

        Args:
            **assignments: assigning the instance value (literal) to the
                existing variables in the Sequence

        """
        display_ir(self, assignments)


@dataclass(frozen=True)
class NamedSequence(SequenceExpr):
    name: str
    sequence: SequenceExpr

    __hash__ = SequenceExpr.__hash__

    @cached_property
    def duration(self) -> Scalar:
        return self.sequence.duration

    def children(self):
        return OrderedDict([("name", self.name), ("sequence", self.sequence)])

    def print_node(self):
        return "NamedSequence"

    def _get_data(self, **assignment):
        return self.name, self.sequence.value

    def figure(self, **assignments):
        return get_ir_figure(self, **assignments)

    def show(self, **assignments):
        display_ir(self, assignments)


@dataclass(frozen=True)
class Slice(SliceTrait, SequenceExpr):
    sequence: SequenceExpr
    interval: Interval

    __hash__ = SequenceExpr.__hash__

    @property
    def _sub_expr(self):
        return self.sequence

    def children(self):
        return [self.interval, self.sequence]

    def print_node(self):
        return "Slice"



from collections import OrderedDict, namedtuple

from bloqade.ir.routine.base import RoutineBase, __pydantic_dataclass_config__
from bloqade.builder.typing import LiteralType
from bloqade.task.batch import LocalBatch
from beartype import beartype
from beartype.typing import Optional, Tuple, Callable, Dict, Any, List, NamedTuple
from pydantic.dataclasses import dataclass
import numpy as np

from bloqade.emulate.codegen.hamiltonian import CompileCache, RydbergHamiltonianCodeGen
from bloqade.emulate.ir.state_vector import AnalogGate, RydbergHamiltonian, StateVector
import traceback


@dataclass(frozen=True, config=__pydantic_dataclass_config__)
class BloqadeServiceOptions(RoutineBase):
    def python(self):
        return BloqadePythonRoutine(self.source, self.circuit, self.params)


@dataclass(frozen=True, config=__pydantic_dataclass_config__)
class BloqadePythonRoutine(RoutineBase):
    @staticmethod
    def process_tasks(runner, tasks, results):
        while not tasks.empty():
            try:
                task_id, (emulator_ir, metadata) = tasks.get()
                result = runner.run_task(emulator_ir, metadata)
                results.put((task_id, result))
            except BaseException as e:
                results.put((task_id, e))

    @dataclass(config=__pydantic_dataclass_config__)
    class EmuRunner:
        compile_cache: Optional[CompileCache]
        solver_args: Dict
        callback: Callable
        callback_args: Tuple

        def run_task(self, emulator_ir, metadata_dict):
            hamiltonian = RydbergHamiltonianCodeGen(
                compile_cache=self.compile_cache
            ).emit(emulator_ir)

            MetaData = namedtuple("MetaData", metadata_dict.keys())
            metadata = MetaData(**{k: float(v) for k, v in metadata_dict.items()})

            zero_state = hamiltonian.space.zero_state(np.complex128)
            (register_data,) = AnalogGate(hamiltonian).apply(
                zero_state, **self.solver_args
            )
            wrapped_register = StateVector(register_data, hamiltonian.space)

            return self.callback(
                wrapped_register, metadata, hamiltonian, *self.callback_args
            )

    def _generate_ir(self, args, blockade_radius, waveform_runtime):
        from bloqade.compiler.passes.emulator import (
            flatten,
            assign,
            generate_emulator_ir,
        )

        params = self.params
        circuit = flatten(self.circuit)

        for task_number, batch_param in enumerate(params.batch_assignments(*args)):
            assignment = {**params.static_params, **batch_param}
            metadata, final_circuit = assign(assignment, circuit)
            emulator_ir = generate_emulator_ir(
                final_circuit, blockade_radius, waveform_runtime
            )
            yield task_number, emulator_ir, metadata

    def _compile(
        self,
        shots: int,
        args: Tuple[LiteralType, ...] = (),
        name: Optional[str] = None,
        blockade_radius: LiteralType = 0.0,
        cache_matrices: bool = False,
        waveform_runtime: str = "interpret",
    ) -> LocalBatch:
        from bloqade.task.bloqade import BloqadeTask

        if cache_matrices:
            matrix_cache = CompileCache()
        else:
            matrix_cache = None

        tasks = OrderedDict()
        it_iter = self._generate_ir(args, blockade_radius, waveform_runtime)
        for task_number, emulator_ir, metadata in it_iter:
            tasks[task_number] = BloqadeTask(shots, emulator_ir, metadata, matrix_cache)

        return LocalBatch(self.source, tasks, name)

    @beartype
    def run(
        self,
        shots: int,
        args: Tuple[LiteralType, ...] = (),
        name: Optional[str] = None,
        blockade_radius: float = 0.0,
        waveform_runtime: str = "interpret",
        interaction_picture: bool = False,
        cache_matrices: bool = False,
        multiprocessing: bool = False,
        num_workers: Optional[int] = None,
        solver_name: str = "dop853",
        atol: float = 1e-7,
        rtol: float = 1e-14,
        nsteps: int = 2_147_483_647,
    ) -> LocalBatch:
        """Run the current program using bloqade python backend

        Args:
            shots (int): number of shots after running state vector simulation
            args (Tuple[LiteralType, ...], optional): The values for parameters defined
            in `args`. Defaults to ().
            name (Optional[str], optional): Name to give this run. Defaults to None.
            blockade_radius (float, optional): Use the Blockade subspace given a
            particular radius. Defaults to 0.0.
            waveform_runtime: (bool, optional): Use Numba to compile the waveforms,
            Defaults to False.
            interaction_picture (bool, optional): Use the interaction picture when
            solving schrodinger equation. Defaults to False.
            cache_matrices (bool, optional): Reuse previously evaluated matrcies when
            possible. Defaults to False.
            multiprocessing (bool, optional): Use multiple processes to process the
            batches. Defaults to False.
            num_workers (Optional[int], optional): Number of processes to run with
            multiprocessing. Defaults to None.
            solver_name (str, optional): Which SciPy Solver to use. Defaults to
            "dop853".
            atol (float, optional): Absolute tolerance for ODE solver. Defaults to
            1e-14.
            rtol (float, optional): Relative tolerance for adaptive step in ODE solver.
            Defaults to 1e-7.
            nsteps (int, optional): Maximum number of steps allowed per integration
            step. Defaults to 2_147_483_647, the maximum value.

        Raises:
            ValueError: Cannot use multiprocessing and cache_matrices at the same time.

        Returns:
            LocalBatch: Batch of local tasks that have been executed.
        """
        if multiprocessing and cache_matrices:
            raise ValueError(
                "Cannot use multiprocessing and cache_matrices at the same time."
            )

        compile_options = dict(
            shots=shots,
            args=args,
            name=name,
            blockade_radius=blockade_radius,
            cache_matrices=cache_matrices,
            waveform_runtime=waveform_runtime,
        )

        solver_options = dict(
            multiprocessing=multiprocessing,
            num_workers=num_workers,
            solver_name=solver_name,
            atol=atol,
            rtol=rtol,
            nsteps=nsteps,
            interaction_picture=interaction_picture,
        )

        batch = self._compile(**compile_options)
        batch._run(**solver_options)

        return batch

    @beartype
    def __call__(
        self,
        *args: LiteralType,
        shots: int = 1,
        name: Optional[str] = None,
        blockade_radius: float = 0.0,
        waveform_runtime: str = "interpret",
        interaction_picture: bool = False,
        multiprocessing: bool = False,
        num_workers: Optional[int] = None,
        cache_matrices: bool = False,
        solver_name: str = "dop853",
        atol: float = 1e-7,
        rtol: float = 1e-14,
        nsteps: int = 2_147_483_647,
    ) -> LocalBatch:
        options = dict(
            shots=shots,
            args=args,
            name=name,
            blockade_radius=blockade_radius,
            waveform_runtime=waveform_runtime,
            multiprocessing=multiprocessing,
            num_workers=num_workers,
            cache_matrices=cache_matrices,
            solver_name=solver_name,
            atol=atol,
            rtol=rtol,
            nsteps=nsteps,
            interaction_picture=interaction_picture,
        )
        return self.run(**options)

    @beartype
    def run_callback(
        self,
        callback: Callable[[StateVector, NamedTuple, RydbergHamiltonian, Any], Any],
        program_args: Tuple[LiteralType, ...] = (),
        callback_args: Tuple = (),
        ignore_exceptions: bool = False,
        blockade_radius: float = 0.0,
        waveform_runtime: str = "interpret",
        interaction_picture: bool = False,
        cache_matrices: bool = False,
        multiprocessing: bool = False,
        num_workers: Optional[int] = None,
        solver_name: str = "dop853",
        atol: float = 1e-7,
        rtol: float = 1e-14,
        nsteps: int = 2_147_483_647,
    ) -> List:
        """Run state-vector simulation with a callback to access full state-vector from
        emulator

        Args:
            callback (Callable[[StateVector, Metadata, RydbergHamiltonian, Any], Any]):
            The callback function to run for each task in batch. See note below for more
            details about the signature of the function.
            program_args (Tuple[LiteralType, ...], optional): The values for parameters
            defined in `args`. Defaults to ().
            callback_args (Tuple[Any,...], optional): Extra arguments to pass into
            ignore_exceptions: (bool, optional) If `True` any exception raised during
            a task will be saved instead of the resulting output of the callback,
            otherwise the first exception by task number will be raised after *all*
            tasks have executed. Defaults to False.
            blockade_radius (float, optional): Use the Blockade subspace given a
            particular radius. Defaults to 0.0.
            waveform_runtime: (str, optional): Specify which runtime to use for
            waveforms. Defaults to "interpret".
            interaction_picture (bool, optional): Use the interaction picture when
            solving schrodinger equation. Defaults to False.
            cache_matrices (bool, optional): Reuse previously evaluated matrcies when
            possible. Defaults to False.
            multiprocessing (bool, optional): Use multiple processes to process the
            batches. Defaults to False.
            num_workers (Optional[int], optional): Number of processes to run with
            multiprocessing. Defaults to None.
            solver_name (str, optional): Which SciPy Solver to use. Defaults to
            "dop853".
            atol (float, optional): Absolute tolerance for ODE solver. Defaults to
            1e-14.
            rtol (float, optional): Relative tolerance for adaptive step in ODE solver.
            Defaults to 1e-7.
            nsteps (int, optional): Maximum number of steps allowed per integration
            step. Defaults to 2_147_483_647, the maximum value.

        Returns:
            List: List of resulting outputs from the callbacks

        Raises:
            RuntimeError: Raises the first error that occurs, only if
            `ignore_exceptions=False`.

        Note:
            For the `callback` function, first argument is the many-body wavefunction
            as a 1D complex numpy array, the second argument is of type `Metadata` which
            is a Named Tuple where the fields correspond to the parameters of that given
            task, RydbergHamiltonian is the object that contains the Hamiltonian used to
            generate the evolution for that task, Finally any optional positional
            arguments are allowed after that. The return value can be anything, the
            results will be collected in a list for each task in the batch.


        """
        if multiprocessing:
            from multiprocessing import Process, Queue, cpu_count
        else:
            from queue import Queue

        if cache_matrices:
            compile_cache = CompileCache()
        else:
            compile_cache = None

        solver_args = dict(
            solver_name=solver_name,
            atol=atol,
            rtol=rtol,
            nsteps=nsteps,
            interaction_picture=interaction_picture,
        )

        runner = self.EmuRunner(
            compile_cache=compile_cache,
            solver_args=solver_args,
            callback=callback,
            callback_args=callback_args,
        )

        tasks = Queue()
        results = Queue()

        total_tasks = 0
        it_iter = self._generate_ir(program_args, blockade_radius, waveform_runtime)
        for task_number, emulator_ir, metadata in it_iter:
            total_tasks += 1
            tasks.put((task_number, (emulator_ir, metadata)))

        workers = []
        if multiprocessing:
            num_workers = max(int(num_workers or cpu_count()), 1)
            num_workers = min(total_tasks, num_workers)

            for _ in range(num_workers):
                worker = Process(
                    target=BloqadePythonRoutine.process_tasks,
                    args=(runner, tasks, results),
                )
                worker.start()

                workers.append(worker)
        else:
            self.process_tasks(runner, tasks, results)

        # blocks until all
        # results have been fetched
        # from the id_results Queue
        id_results = []
        for i in range(total_tasks):
            id_results.append(results.get())

        if workers:
            for worker in workers:
                worker.join()

            tasks.close()
            results.close()

        id_results.sort(key=lambda x: x[0])
        results = []

        for task_id, result in id_results:
            if not ignore_exceptions and isinstance(result, BaseException):
                try:
                    raise result
                except BaseException:
                    raise RuntimeError(
                        f"{result.__class__.__name__} occured during child process "
                        f"running for task number {task_id}:\n{traceback.format_exc()}"
                    )

            results.append(result)

        return results



from bloqade.ir.location import location
from bloqade.ir import analog_circuit
from bloqade.submission.ir.parallel import ClusterLocationInfo, ParallelDecoder
from bloqade.submission.capabilities import QuEraCapabilities
from bloqade.ir.visitor import BloqadeIRVisitor
from beartype.typing import Optional
import numpy as np
from decimal import Decimal
from pydantic.dataclasses import dataclass
from beartype.typing import List, Tuple


@dataclass
class AHSLatticeData:
    sites: List[Tuple[Decimal, Decimal]]
    filling: List[int]
    parallel_decoder: Optional[ParallelDecoder]


class GenerateLattice(BloqadeIRVisitor):
    def __init__(self, capabilities: Optional[QuEraCapabilities] = None):
        self.capabilities = capabilities
        self.parallel_decoder = None

    def generic_visit(self, node):
        # dispatch all AtomArrangement nodes to visit_register
        # otherwise dispatch to super
        if isinstance(node, location.AtomArrangement):
            self.visit_register(node)

        super().generic_visit(node)

    def visit_register(self, node: location.AtomArrangement):
        # default visitor for AtomArrangement
        self.sites = []
        self.filling = []

        for location_info in node.enumerate():
            site = tuple(ele() for ele in location_info.position)
            self.sites.append(site)
            self.filling.append(location_info.filling.value)

    def visit_location_ParallelRegister(self, node: location.ParallelRegister):
        from bloqade.ir.location.location import ParallelRegisterInfo

        info = ParallelRegisterInfo(node)

        if self.capabilities is None:
            raise ValueError("Cannot parallelize register without device capabilities.")

        height_max = self.capabilities.capabilities.lattice.area.height / Decimal(
            "1e-6"
        )
        width_max = self.capabilities.capabilities.lattice.area.width / Decimal("1e-6")
        number_sites_max = (
            self.capabilities.capabilities.lattice.geometry.number_sites_max
        )

        register_filling = np.asarray(info.register_filling)

        register_locations = np.asarray(
            [[s() for s in location] for location in info.register_locations]
        )
        register_locations = register_locations - register_locations.min(axis=0)

        shift_vectors = np.asarray(
            [[s() for s in shift_vector] for shift_vector in info.shift_vectors]
        )

        # build register by stack method because
        # shift_vectors might not be rectangular
        c_stack = [(0, 0)]
        visited = set([(0, 0)])
        mapping = []
        global_site_index = 0
        sites = []
        filling = []
        while c_stack:
            if len(mapping) + len(info.register_locations) > number_sites_max:
                break

            cluster_index = c_stack.pop()

            shift = (
                shift_vectors[0] * cluster_index[0]
                + shift_vectors[1] * cluster_index[1]
            )

            new_register_locations = register_locations + shift

            # skip clusters that fall out of bounds
            if (
                np.any(new_register_locations < 0)
                or np.any(new_register_locations[:, 0] > width_max)
                or np.any(new_register_locations[:, 1] > height_max)
            ):
                continue

            new_cluster_indices = [
                (cluster_index[0] + 1, cluster_index[1]),
                (cluster_index[0], cluster_index[1] + 1),
                (cluster_index[0] - 1, cluster_index[1]),
                (cluster_index[0], cluster_index[1] - 1),
            ]

            for new_cluster_index in new_cluster_indices:
                if new_cluster_index not in visited:
                    visited.add(new_cluster_index)
                    c_stack.append(new_cluster_index)

            for cluster_location_index, (loc, filled) in enumerate(
                zip(new_register_locations[:], register_filling)
            ):
                site = tuple(loc)
                sites.append(site)
                filling.append(filled)

                mapping.append(
                    ClusterLocationInfo(
                        cluster_index=cluster_index,
                        global_location_index=global_site_index,
                        cluster_location_index=cluster_location_index,
                    )
                )

                global_site_index += 1

        self.sites = sites
        self.filling = filling
        self.parallel_decoder = ParallelDecoder(mapping=mapping)

    def visit_analog_circuit_AnalogCircuit(self, node: analog_circuit.AnalogCircuit):
        self.visit(node.register)

    def emit(self, node) -> AHSLatticeData:
        self.visit(node)

        return AHSLatticeData(
            sites=self.sites,
            filling=self.filling,
            parallel_decoder=self.parallel_decoder,
        )



from bloqade.ir.visitor import BloqadeIRVisitor
import bloqade.ir.control.waveform as waveform
from beartype.typing import Dict, FrozenSet, Set
from beartype import beartype
from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class WaveformScanResult:
    bindings: Dict[waveform.Waveform, str]
    imports: Dict[str, FrozenSet[str]]


class WaveformScan(BloqadeIRVisitor):
    def __init__(
        self,
        bindings: Dict[waveform.Waveform, str] = {},
        imports: Dict[str, Set[str]] = {},
    ):
        self.bindings = dict(bindings)
        self.imports = dict(imports)
        self.i = 0

    def add_binding(self, node: waveform.Waveform):
        if node in self.bindings:
            return

        symbol = f"__bloqade_var{self.i}"

        while symbol in self.imports.values():
            self.i += 1
            symbol = f"__bloqade_var{self.i}"

        self.i += 1

        self.bindings[node] = symbol

    def generic_visit(self, node: waveform.Waveform):
        if isinstance(node, waveform.Waveform):
            self.add_binding(node)
        super().generic_visit(node)

    @beartype
    def scan(self, node: waveform.Waveform) -> WaveformScanResult:
        self.visit(node)
        imports = {
            module: frozenset(imports) for module, imports in self.imports.items()
        }
        return WaveformScanResult(self.bindings, imports)



import bloqade.ir.location as location
import bloqade.ir.control.sequence as sequence
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.field as field
import bloqade.ir.analog_circuit as analog_circuit
import bloqade.ir.control.waveform as waveform
from bloqade.ir.location import (
    Chain,
    Honeycomb,
    Kagome,
    Lieb,
    ListOfLocations,
    Rectangular,
    Square,
    Triangular,
)
from bloqade.ir.location.location import LocationInfo, ParallelRegister
import bloqade.ir.scalar as scalar

from bloqade.ir.visitor import BloqadeIRVisitor, BloqadeNodeTypes

import json

from typing import Any, Dict


class BloqadeIRSerializer(json.JSONEncoder, BloqadeIRVisitor):
    def visit_scalar_Literal(self, node: scalar.Literal) -> Dict[str, Dict[str, str]]:
        return {"literal": {"value": str(node.value)}}

    def visit_scalar_AssignedVariable(
        self, node: scalar.AssignedVariable
    ) -> Dict[str, Any]:
        return {
            "assigned_variable": {
                "name": node.name,
                "value": str(node.value),
            }
        }

    def visit_scalar_Negative(self, node: scalar.Negative) -> Dict[str, Any]:
        return {"negative": {"expr": self.visit(node.expr)}}

    def visit_scalar_Add(self, node: scalar.Add) -> Dict[str, Any]:
        return {"add": {"lhs": self.visit(node.lhs), "rhs": self.visit(node.rhs)}}

    def visit_scalar_Mul(self, node: scalar.Mul) -> Dict[str, Any]:
        return {"mul": {"lhs": self.visit(node.lhs), "rhs": self.visit(node.rhs)}}

    def visit_scalar_Div(self, node: scalar.Div) -> Dict[str, Any]:
        return {"div": {"lhs": self.visit(node.lhs), "rhs": self.visit(node.rhs)}}

    def visit_scalar_Min(self, node: scalar.Min) -> Dict[str, Any]:
        return {"min": {"exprs": list(map(self.visit, node.exprs))}}

    def visit_scalar_Max(self, node: scalar.Max) -> Dict[str, Any]:
        return {"max": {"exprs": list(map(self.visit, node.exprs))}}

    def visit_scalar_Slice(self, node: scalar.Slice):
        return {
            "slice_scalar": {
                "expr": self.visit(node.expr),
                "interval": self.visit(node.interval),
            }
        }

    def visit_scalar_Interval(self, node: scalar.Interval) -> Dict[str, Any]:
        return {
            "interval": {
                "start": self.visit(node.start) if node.start is not None else None,
                "stop": self.visit(node.stop) if node.stop is not None else None,
            }
        }

    def visit_waveform_Constant(self, node: waveform.Constant) -> Dict[str, Any]:
        return {
            "constant": {
                "value": self.visit(node.value),
                "duration": self.visit(node.duration),
            }
        }

    def visit_waveform_Linear(self, node: waveform.Linear) -> Dict[str, Any]:
        return {
            "linear": {
                "start": self.visit(node.start),
                "stop": self.visit(node.stop),
                "duration": self.visit(node.duration),
            }
        }

    def visit_waveform_Poly(self, node: waveform.Poly) -> Dict[str, Any]:
        return {
            "poly": {
                "coeffs": list(map(self.visit, node.coeffs)),
                "duration": self.visit(node.duration),
            }
        }

    def visit_waveform_PythonFn(self, node: waveform.PythonFn) -> Dict[str, Any]:
        raise ValueError("Bloqade does not support serialization of Python code.")

    def visit_waveform_Negative(self, node: waveform.Negative) -> Dict[str, Any]:
        return {"negative_waveform": {"waveform": self.visit(node.waveform)}}

    def visit_waveform_Add(self, node: waveform.Add) -> Dict[str, Any]:
        return {
            "add_waveform": {
                "left": self.visit(node.left),
                "right": self.visit(node.right),
            }
        }

    def visit_waveform_Scale(self, node: waveform.Scale) -> Dict[str, Any]:
        return {
            "scale": {
                "waveform": self.visit(node.waveform),
                "scalar": self.visit(node.scalar),
            }
        }

    def visit_waveform_Slice(self, node: waveform.Slice) -> Dict[str, Any]:
        return {
            "slice_waveform": {
                "waveform": self.visit(node.waveform),
                "interval": self.visit(node.interval),
            }
        }

    def visit_waveform_Sample(self, node: waveform.Sample) -> Dict[str, Any]:
        return {
            "sample": {
                "waveform": self.visit(node.waveform),
                "dt": self.visit(node.dt),
                "interpolation": node.interpolation.value,
            }
        }

    def visit_waveform_Append(self, node: waveform.Append) -> Dict[str, Any]:
        return {"append_waveform": {"waveforms": list(map(self.visit, node.waveforms))}}

    def visit_waveform_Record(self, node: waveform.Record) -> Dict[str, Any]:
        return {
            "record": {
                "var": self.visit(node.var),
                "waveform": self.visit(node.waveform),
            }
        }

    def visit_waveform_Smooth(self, node: waveform.Smooth) -> Dict[str, Any]:
        return {
            "smooth": {
                "waveform": self.visit(node.waveform),
                "radius": self.visit(node.radius),
                "kernel": type(node.kernel).__name__,
            }
        }

    def visit_waveform_AllignedWaveform(
        self, node: waveform.AlignedWaveform
    ) -> Dict[str, Any]:
        if isinstance(node.value, scalar.Scalar):
            value = self.visit(node.value)
        else:
            value = node.value.value
        return {
            "alligned": {
                "waveform": self.visit(node.waveform),
                "allignment": self.visit(node.allignment),
                "value": value,
            }
        }

    def visit_bravais_Chain(self, node: Chain) -> Any:
        return {
            "chain": {
                "lattice_spacing": self.visit(node.lattice_spacing),
                "L": node.L,
                "vertical_chain": node.vertical_chain,
            }
        }

    def visit_bravais_Honeycomb(self, node: Honeycomb) -> Any:
        return {
            "honeycomb": {
                "lattice_spacing": self.visit(node.lattice_spacing),
                "L1": node.shape[0],
                "L2": node.shape[1],
            }
        }

    def visit_bravais_Kagome(self, node: Kagome) -> Any:
        return {
            "kagome": {
                "lattice_spacing": self.visit(node.lattice_spacing),
                "L1": node.shape[0],
                "L2": node.shape[1],
            }
        }

    def visit_bravais_Lieb(self, node: Lieb) -> Any:
        return {
            "lieb": {
                "lattice_spacing": self.visit(node.lattice_spacing),
                "L1": node.shape[0],
                "L2": node.shape[1],
            }
        }

    def visit_locaiton_LocationInfo(self, node: LocationInfo) -> Any:
        return {
            "location_info": {
                "position": [[self.visit(x), self.visit(y)] for x, y in node.position],
                "filling": node.filling.value,
            }
        }

    def visit_location_ListOfLocations(self, node: ListOfLocations) -> Any:
        return {
            "list_of_locations": {
                "location_list": list(map(self.visit, node.location_list))
            }
        }

    def visit_bravais_Rectangular(self, node: Rectangular) -> Any:
        return {
            "rectangular": {
                "lattice_spacing_x": self.visit(node.lattice_spacing_x),
                "lattice_spacing_y": self.visit(node.lattice_spacing_y),
                "width": node.width,
                "height": node.height,
            }
        }

    def visit_bravais_Square(self, node: Square) -> Any:
        return {
            "square": {
                "lattice_spacing": self.visit(node.lattice_spacing),
                "L1": node.shape[0],
                "L2": node.shape[1],
            }
        }

    def visit_bravais_Triangular(self, node: Triangular) -> Any:
        return {
            "triangular": {
                "lattice_spacing": self.visit(node.lattice_spacing),
                "L1": node.shape[0],
                "L2": node.shape[1],
            }
        }

    def visit_location_ParallelRegister(self, node: ParallelRegister) -> Any:
        return {
            "parallel_register": {
                "register": self.visit(node.register),
                "cluster_spacing": self.visit(node.cluster_spacing),
            }
        }

    def visit_sequence_Sequence(self, node: sequence.Sequence) -> Dict[str, Any]:
        object_to_str = {
            sequence.rydberg: "rydberg",
            sequence.hyperfine: "hyperfine",
        }
        return {
            "sequence": {
                "pulses": {
                    object_to_str[k]: self.visit(v) for k, v in node.pulses.items()
                }
            }
        }

    def visit_sequence_NamedSequence(
        self, node: sequence.NamedSequence
    ) -> Dict[str, Any]:
        return {
            "named_sequence": {
                "name": node.name,
                "sub_sequence": self.visit(node.sub_sequence),
            }
        }

    def visit_sequence_Append(self, node: sequence.Append) -> Dict[str, Any]:
        return {
            "append_sequence": {"sequences": [self.visit(s) for s in node.sequences]}
        }

    def visit_sequence_Slice(self, node: sequence.Slice) -> Dict[str, Any]:
        return {
            "slice_sequence": {
                "sequence": self.visit(node.sequence),
                "interval": self.visit(node.interval),
            }
        }

    def visit_pulse_Pulse(self, node: pulse.Pulse):
        object_map_str = {
            pulse.detuning: "detuning",
            pulse.rabi.amplitude: "rabi_frequency_amplitude",
            pulse.rabi.phase: "rabi_frequency_phase",
        }
        return {
            "pulse": {
                "fields": {
                    object_map_str[k]: self.visit(v) for k, v in node.fields.items()
                }
            }
        }

    def visit_pulse_Append(self, node: pulse.Append) -> Dict[str, Any]:
        return {"append_pulse": {"pulses": [self.visit(p) for p in node.pulses]}}

    def visit_pulse_NamedPulse(self, node: pulse.NamedPulse) -> Dict[str, Any]:
        return {
            "named_pulse": {"name": node.name, "sub_pulse": self.visit(node.sub_pulse)}
        }

    def visit_pulse_Slice(self, node: pulse.Slice) -> Dict[str, Any]:
        return {
            "slice_pulse": {
                "pulse": self.visit(node.sub_pulse),
                "interval": self.visit(node.interval),
            }
        }

    def visit_field_Location(self, node: field.Location) -> Dict[str, Any]:
        return {"location": {"label": node.label}}

    def visit_field_ScaledLocations(
        self, node: field.ScaledLocations
    ) -> Dict[str, Any]:
        return {
            "scaled_locations": [
                (self.visit(k), self.visit(v)) for k, v in node.items()
            ]
        }

    def visit_field_Uniform(self, node: field.Uniform) -> Dict[str, Any]:
        return {"uniform": {}}

    def visit_field_RunTimeVector(self, node: field.RunTimeVector) -> Dict[str, Any]:
        return {"run_time_vector": {"name": node.name}}

    def visit_field_AssignedRunTimeVector(
        self, node: field.AssignedRunTimeVector
    ) -> Dict[str, Any]:
        return {
            "assigned_run_time_vector": {
                "name": node.name,
                "value": [str(v) for v in node.value],
            }
        }

    def visit_field_Field(self, node: field.Field) -> Any:
        return {
            "field": {
                "value": [
                    (self.visit(k), self.visit(v)) for k, v in node.drives.items()
                ]
            }
        }

    def visit_analog_circuit_AnalogCircuit(
        self, node: analog_circuit.AnalogCircuit
    ) -> Any:
        return {
            "analog_circuit": {
                "sequence": self.visit(node.sequence),
                "register": self.visit(node.register),
                "static_params": node.static_params,
                "batch_params": node.batch_params,
                "order": node.order,
            }
        }

    def default(self, obj: Any) -> Any:
        if isinstance(obj, BloqadeNodeTypes):
            return self.visit(obj)
        else:
            return super().default(obj)


class BloqadeIRDeserializer:
    constructors = {
        "literal": scalar.Literal,
        "variable": scalar.Variable,
        "assigned_variable": scalar.AssignedVariable,
        "negative": scalar.Negative,
        "add": scalar.Add,
        "mul": scalar.Mul,
        "div": scalar.Div,
        "min": scalar.Min,
        "max": scalar.Max,
        "slice_scalar": scalar.Slice,
        "interval": scalar.Interval,
        "constant": waveform.Constant,
        "linear": waveform.Linear,
        "poly": waveform.Poly,
        "negative_waveform": waveform.Negative,
        "add_waveform": waveform.Add,
        "scale": waveform.Scale,
        "slice_waveform": waveform.Slice,
        "sample": waveform.Sample,
        "append_waveform": waveform.Append,
        "record": waveform.Record,
        "smooth": waveform.Smooth,
        "alligned": waveform.AlignedWaveform,
        "chain": location.Chain,
        "honeycomb": location.Honeycomb,
        "kagome": location.Kagome,
        "lieb": location.Lieb,
        "list_of_locations": location.ListOfLocations,
        "rectangular": location.Rectangular,
        "square": location.Square,
        "triangular": location.Triangular,
        "parallel_register": location.ParallelRegister,
        "sequence": sequence.Sequence,
        "named_sequence": sequence.NamedSequence,
        "append_sequence": sequence.Append,
        "slice_sequence": sequence.Slice,
        "pulse": pulse.Pulse,
        "append_pulse": pulse.Append,
        "named_pulse": pulse.NamedPulse,
        "slice_pulse": pulse.Slice,
        "location": field.Location,
        "scaled_locations": field.ScaledLocations,
        "uniform": field.Uniform,
        "run_time_vector": field.RunTimeVector,
        "assigned_run_time_vector": field.AssignedRunTimeVector,
        "field": field.Field,
        "analog_circuit": analog_circuit.AnalogCircuit,
    }

    @classmethod
    def object_hook(cls, obj):
        if isinstance(obj, dict) and len(obj) == 1:
            ((head, options),) = obj.items()
            if head in cls.constructors:
                return cls.constructors[head](**options)

        return obj



from beartype import beartype
from decimal import Decimal
from beartype.typing import Optional
from bloqade.ir.visitor import BloqadeIRVisitor
from bloqade.ir.control import field, pulse
from bloqade.submission.ir.parallel import ParallelDecoder
from bloqade.ir import analog_circuit, scalar


class GenerateLatticeSiteCoefficients(BloqadeIRVisitor):
    def __init__(self, parallel_decoder: Optional[ParallelDecoder] = None):
        self.n_sites = None
        self.parallel_decoder = parallel_decoder
        self.lattice_site_coefficients = None

    def post_spatial_modulation_visit(self):
        if self.parallel_decoder is None:
            # if we are not parallelizing, we don't need to do anything
            return

        # create a copy of the cluster site coefficients
        lattice_site_coefficients = list(self.lattice_site_coefficients)
        # insert the cluster site coefficients into the parallelized
        # lattice site coefficients
        self.lattice_site_coefficients = []
        # sort by global location index so that we can insert the
        # cluster site coefficients in the correct order
        sorted_decoder_mapping = sorted(
            self.parallel_decoder.mapping, key=lambda x: x.global_location_index
        )
        for cluster_site_info in sorted_decoder_mapping:
            self.lattice_site_coefficients.append(
                lattice_site_coefficients[cluster_site_info.cluster_location_index]
            )

    # We don't need to visit UniformModulation because local detuning
    # UniformModulation is merged into global detuning

    def visit_field_ScaledLocations(self, node: field.ScaledLocations):
        self.lattice_site_coefficients = []

        for i in range(self.n_sites):
            value = node.value.get(field.Location(i), scalar.Literal(0))
            self.lattice_site_coefficients.append(value())

    def visit_field_AssignedRunTimeVector(self, node: field.AssignedRunTimeVector):
        self.lattice_site_coefficients = list(map(Decimal, map(str, node.value)))

    def visit_field_Field(self, node: field.Field):
        (local_detuning_spatial_modulation,) = node.drives.keys() - {field.Uniform}
        self.visit(local_detuning_spatial_modulation)
        self.post_spatial_modulation_visit()

    def visit_pulse_Pulse(self, node: pulse.Pulse):
        self.visit(node.fields[pulse.detuning])

    def visit_analog_circuit_AnalogCircuit(self, node: analog_circuit.AnalogCircuit):
        self.n_sites = node.register.n_sites
        self.visit(node.sequence)

    # needs to be an instance of analog_circuit.AnalogCircuit
    # because we need to know the number of sites
    @beartype
    def emit(self, node: analog_circuit.AnalogCircuit):
        self.visit(node)
        return self.lattice_site_coefficients



# from juliacall import Main, TypeValue  # type: ignore
# from typing import List, Dict, Any
# from dataclasses import dataclass
# from .to_julia import ToJulia

# PythonCall = Main.seval("PythonCall")


# @dataclass(frozen=True)
# class JLTypeVar:
#     name: str


# class JLType:
#     """Superclass for all Julia types."""

#     def __init__(self, expr, *typeparams) -> None:
#         self.expr = expr
#         self.obj = getattr(Main, expr)
#         params = []
#         for param in typeparams:
#             if isinstance(param, JLType):
#                 params.append(param.obj)
#             elif isinstance(param, TypeValue):
#                 params.append(param)
#             else:
#                 raise Exception(f"Unknown type parameter: {param}")
#         self.typeparams = params

#     def __call__(self, x) -> Any:
#         if self.typeparams:
#             jl_type = self.obj[self.typeparams]
#         else:
#             jl_type = self.obj
#         return PythonCall.pyconvert(jl_type, x)

#     def __getitem__(self, *typeparams: Any) -> Any:
#         return JLType(self.expr, *typeparams)

#     def __repr__(self) -> str:
#         repr = self.expr
#         if self.typeparams:
#             repr += "[" + ",".join(map(str, self.typeparams)) + "]"
#         return repr


# class JLVectorType(JLType):
#     def __init__(self, expr, *typeparams) -> None:
#         super().__init__(expr, *typeparams)

#     def __call__(self, x: List[ToJulia]) -> Any:
#         type = self.obj[self.typeparams]
#         ret = type(Main.undef, len(x))
#         for i, v in enumerate(x):
#             ret[i] = v.julia()
#         return ret

#     def __getitem__(self, typeparams: Any) -> Any:
#         assert len(typeparams) == 1
#         assert self.typeparams == []
#         return JLVectorType(*typeparams)


# class JLDictType(JLType):
#     def __init__(self, expr, *typeparams) -> None:
#         super().__init__(expr, *typeparams)

#     def __call__(self, x: Dict[ToJulia, ToJulia]) -> Any:
#         type = self.obj[self.typeparams]
#         ret = type()
#         for k, v in x.items():
#             ret[k.julia()] = v.julia()
#         return ret

#     def __getitem__(self, *typeparams: Any) -> Any:
#         assert len(self.typeparams) + len(typeparams) == 2
#         return JLDictType(self.expr, *typeparams)


# Int32 = JLType("Int32")
# Int64 = JLType("Int64")
# Float32 = JLType("Float32")
# Float64 = JLType("Float64")
# Complex = JLType("Complex")
# ComplexF32 = JLType("ComplexF32")
# ComplexF64 = JLType("ComplexF64")
# String = JLType("String")
# Bool = JLType("Bool")
# Symbol = JLType("Symbol")
# Vector = JLVectorType("Vector")
# Dict = JLDictType("Dict")



from functools import reduce
from bloqade.ir.visitor import BloqadeIRTransformer
import bloqade.ir.control.waveform as waveform
import numpy as np


def concat(lhs: waveform.Waveform, rhs: waveform.Waveform) -> waveform.Waveform:
    return lhs.append(rhs)


class NormalizeWaveformPython(BloqadeIRTransformer):
    def visit_waveform_Sample(self, node: waveform.Sample):
        times, values = node.samples()

        durations = np.diff(times)

        if node.interpolation is waveform.Interpolation.Constant:
            segments = [waveform.Constant(*args) for args in zip(values, durations)]
        elif node.interpolation is waveform.Interpolation.Linear:
            segments = [
                waveform.Linear(*args)
                for args in zip(values[:-1], values[1:], durations)
            ]

        return reduce(concat, segments)



from .lattice import GenerateLattice
from .lattice_site_coefficients import GenerateLatticeSiteCoefficients
from .piecewise_constant import GeneratePiecewiseConstantChannel, PiecewiseConstant
from .piecewise_linear import GeneratePiecewiseLinearChannel, PiecewiseLinear

__all__ = [
    "GenerateLattice",
    "GenerateLatticeSiteCoefficients",
    "GeneratePiecewiseConstantChannel",
    "GeneratePiecewiseLinearChannel",
    "PiecewiseConstant",
    "PiecewiseLinear",
]



from beartype import beartype
from bloqade.ir import analog_circuit
import bloqade.ir.control.field as field
import bloqade.ir.control.pulse as pulse
import bloqade.ir.control.sequence as sequence

from bloqade.ir.visitor import BloqadeIRVisitor


class ValidateChannels(BloqadeIRVisitor):
    """Checks to make sure the given sequence can be compiled to hardware.

    This check looks at the spatial modulations and the level coupling
    to determine if the sequence can be compiled to hardware.

    """

    def __init__(self):
        self.field_name = None

    def visit_sequence_Sequence(self, node: sequence.Sequence) -> None:
        for level_coupling, pulse_expr in node.pulses.items():
            if level_coupling == sequence.hyperfine:
                raise ValueError("Hyperfine coupling not supported by AHS")

            self.visit(pulse_expr)

    def visit_pulse_Pulse(self, node: pulse.Pulse) -> None:
        for field_name, field_expr in node.fields.items():
            self.field_name = field_name
            self.visit(field_expr)

    def visit_field_Field(self, node: field.Field) -> None:
        non_uniform_modulations = node.drives.keys() - {field.Uniform}
        spatial_modulation_str = "\n\n".join(map(str, non_uniform_modulations))

        if self.field_name in [pulse.rabi.amplitude, pulse.rabi.phase]:
            if non_uniform_modulations:
                raise ValueError(
                    f"{self.field_name} can only have uniform spatial modulation, "
                    f"found:\n\n{spatial_modulation_str}"
                )
        else:
            if len(non_uniform_modulations) > 1:
                raise ValueError(
                    f"{self.field_name} can only have one non-uniform spatial "
                    f"modulation, found:\n\n{spatial_modulation_str}"
                )

            if non_uniform_modulations:
                self.visit(non_uniform_modulations.pop())

    def visit_field_ScaledLocations(self, node: field.ScaledLocations) -> None:
        for loc in node.value.keys():
            if loc.value >= self.n_sites:
                raise ValueError(
                    "location index out of range: "
                    f"found {loc.value} with a total number of sites: {self.n_sites}."
                    f"Error found in the {self.field_name} field "
                    f"with spatial modulation:\n\n{node}"
                )

    def visit_field_AssignedRunTimeVector(
        self, node: field.AssignedRunTimeVector
    ) -> None:
        if len(node.value) != self.n_sites:
            raise ValueError(
                "length of runtime vector does not match number of sites in the "
                f"register. expected {self.n_sites}, found {len(node.value)} for "
                f"the {self.field_name} field with spatial modulation:"
                f"\n\n{node}"
            )

    def visit_analog_circuit_AnalogCircuit(
        self, node: analog_circuit.AnalogCircuit
    ) -> None:
        self.n_sites = node.register.n_sites
        self.visit(node.sequence)

    # needs to be an instance of analog_circuit.AnalogCircuit
    # because we need to know the number of sites
    @beartype
    def scan(self, node: analog_circuit.AnalogCircuit) -> None:
        self.visit(node)



from decimal import Decimal

from beartype import beartype
from bloqade.ir.visitor import BloqadeIRVisitor
import bloqade.ir.control.waveform as waveform
from bloqade.compiler.analysis.python.waveform import WaveformScanResult
from beartype.typing import Optional, Tuple
from random import randint


class CodegenPythonWaveform(BloqadeIRVisitor):
    def __init__(
        self,
        scan_result: WaveformScanResult,
        time_str: str = "time",
        indent_level: int = 0,
        jit_compiled: bool = True,
    ):
        self.jit_compiled = jit_compiled
        self.time_str = time_str
        self.bindings = dict(scan_result.bindings)
        self.imports = dict(scan_result.imports)
        self.exprs = []
        self.head_binding = None
        self.indent_level = indent_level
        self.indent_expr = "    " * (self.indent_level + 1)
        self.indent_func = "    " * self.indent_level

    @staticmethod
    def gen_func_binding():
        func_binding = f"__bloqade_waveform_{randint(0, 2**32)}"
        while func_binding in globals():
            func_binding = f"__bloqade_waveform_{randint(0, 2**32)}"

        return func_binding

    def visit(self, node):
        super().visit(node)
        if isinstance(node, waveform.Waveform):
            self.head_binding = self.bindings[node]

    def visit_waveform_Constant(self, node: waveform.Constant):
        self.exprs.append(f"{self.indent_expr}{self.bindings[node]} = {node.value()}")

    def visit_waveform_Linear(self, node: waveform.Linear):
        slope = (node.stop - node.start) / node.duration

        self.exprs.append(
            f"{self.indent_expr}{self.bindings[node]} = "
            f"{slope()} * ({self.time_str}) + {node.start()}"
        )

    def visit_waveform_Poly(self, node: waveform.Poly):
        coeff_values = [coeff() for coeff in node.coeffs]
        binding = self.bindings[node]

        terms = [str(coeff_values[0])] + [
            f"{coeff} * ({self.time_str}) ** {p}"
            for p, coeff in enumerate(coeff_values[1:], 1)
        ]
        term_sum = " + ".join(terms)
        self.exprs.append(f"{self.indent_expr}{binding} = {term_sum}")

    def visit_waveform_PythonFn(self, node: waveform.PythonFn):
        from numba import njit

        sorted_parameters = sorted(node.parameters, key=lambda p: p.name)

        args = ", ".join(
            [f"{param.name} = {param.value}" for param in sorted_parameters]
        )
        func_binding = self.gen_func_binding()

        globals()[func_binding] = njit(node.fn) if self.jit_compiled else node.fn

        if args:
            self.exprs.append(
                f"{self.indent_expr}{self.bindings[node]} = "
                f"{func_binding}({self.time_str}, {args})"
            )
        else:
            self.exprs.append(
                f"{self.indent_expr}{self.bindings[node]} = "
                f"{func_binding}({self.time_str})"
            )

    def visit_waveform_Add(self, node: waveform.Add):
        self.visit(node.left)
        self.visit(node.right)

        left_duration = node.left.duration()
        right_duration = node.right.duration()

        if left_duration == right_duration:
            self.exprs.append(
                f"{self.indent_expr}{self.bindings[node]} = "
                f"{self.bindings[node.left]} + {self.bindings[node.right]}"
            )
        elif left_duration > right_duration:
            self.exprs.append(
                f"{self.indent_expr}{self.bindings[node]} = "
                f"{self.bindings[node.left]} + {self.bindings[node.right]} "
                f"if {self.time_str} < {right_duration} else {self.bindings[node.left]}"
            )
        else:
            self.exprs.append(
                f"{self.indent_expr}{self.bindings[node]} = "
                f"{self.bindings[node.left]} + {self.bindings[node.right]} "
                f"if {self.time_str} < {left_duration} else {self.bindings[node.right]}"
            )

    def visit_waveform_Negative(self, node: waveform.Negative):
        self.visit(node.waveform)
        self.exprs.append(
            f"{self.indent_expr}{self.bindings[node]} = -{self.bindings[node.waveform]}"
        )

    def visit_waveform_Scale(self, node: waveform.Scale):
        self.visit(node.waveform)
        self.exprs.append(
            f"{self.indent_expr}{self.bindings[node]} = "
            f"{node.scalar()} * {self.bindings[node.waveform]}"
        )

    def visit_waveform_Slice(self, node: waveform.Slice):
        shift = node.interval.start() if node.interval.start else Decimal("0")

        compiler = CodegenPythonWaveform(
            WaveformScanResult(self.bindings, self.imports),
            time_str=f"{self.time_str} + {shift}",
            indent_level=self.indent_level,
        )

        compiler.visit(node.waveform)

        self.exprs.extend(compiler.exprs)
        self.exprs.append(
            f"{self.indent_expr}{self.bindings[node]} = "
            f"{compiler.bindings[node.waveform]}"
        )

    def visit_waveform_Append(self, node: waveform.Append):
        time_shift = Decimal("0")

        wf = node.waveforms[0]

        compiler = CodegenPythonWaveform(
            WaveformScanResult(self.bindings, self.imports),
            time_str=f"{self.time_str} - {time_shift}",
            indent_level=self.indent_level + 1,
        )

        compiler.visit(wf)
        time_shift += wf.duration()
        self.exprs.append(f"{self.indent_expr}if {self.time_str} < {time_shift}:")
        self.exprs.extend(compiler.exprs)
        self.exprs.append(
            f"{compiler.indent_expr}{self.bindings[node]} = {compiler.head_binding}"
        )

        for wf in node.waveforms[1:]:
            compiler = CodegenPythonWaveform(
                WaveformScanResult(self.bindings, self.imports),
                time_str=f"{self.time_str} - {time_shift}",
                indent_level=self.indent_level + 1,
            )

            compiler.visit(wf)
            time_shift += wf.duration()
            self.exprs.append(
                f"{self.indent_expr}elif {self.time_str} <= {time_shift}:"
            )
            self.exprs.extend(compiler.exprs)
            self.exprs.append(
                f"{compiler.indent_expr}{self.bindings[node]} = {compiler.head_binding}"
            )

        self.exprs.append(f"{self.indent_expr}else:")
        self.exprs.append(f"{compiler.indent_expr}{self.bindings[node]} = 0")

    @beartype
    def emit_func(
        self, node: waveform.Waveform, func_binding: Optional[str] = None
    ) -> Tuple[str, str]:
        func_binding = self.gen_func_binding() if func_binding is None else func_binding
        self.visit(node)
        body = "\n".join(self.exprs)
        func = (
            f"def {func_binding}(time):\n"
            f"    if time > {node.duration()}:"
            f"\n        return 0"
            f"\n{body}"
            f"\n    return {self.head_binding}"
        )
        func_code = self.indent_func + func.replace("\n", "\n" + self.indent_func)
        return func_binding, func_code

    @beartype
    def compile(self, node: waveform.Waveform):
        func_binding, func = self.emit_func(node)
        imports = "\n".join(
            [
                f"from {module} import {', '.join(funcs)}"
                for module, funcs in self.imports.items()
            ]
        )

        if self.jit_compiled:
            func = (
                f"{imports}"
                f"\nfrom numba import njit, float64"
                f"\n\n"
                f"@njit(float64(float64))"
                f"\n{func}"
            )
        else:
            func = f"{imports}\n\n{func}"

        globs = globals()
        exec(func, globs)

        return globs[func_binding]



from bloqade.ir.visitor import BloqadeIRVisitor
from bloqade.ir.control import field, pulse, sequence, waveform


class ValidatePiecewiseConstantChannel(BloqadeIRVisitor):
    def __init__(
        self,
        level_coupling: sequence.LevelCoupling,
        field_name: pulse.FieldName,
        spatial_modulations: field.SpatialModulation,
    ):
        self.level_coupling = level_coupling
        self.field_name = field_name
        self.spatial_modulations = spatial_modulations

    def emit_shape_error(self, node: waveform.Waveform):
        return (
            "failed to compile waveform to piecewise constant. "
            f"found non-constant waveform: {node.print_node()} "
            f"for the {self.level_coupling} {self.field_name} with spatial "
            f"modulation:\n{self.spatial_modulations}"
        )

    def visit_waveform_Linear(self, node: waveform.Linear) -> None:
        if node.start() != node.stop():
            raise ValueError(self.emit_shape_error(node))

    def visit_waveform_Poly(self, node: waveform.Poly) -> None:
        if len(node.coeffs) > 1:
            raise ValueError(self.emit_shape_error(node))

    def visit_waveform_Sample(self, node: waveform.Sample) -> None:
        if node.interpolation != waveform.Interpolation.Constant:
            raise ValueError(self.emit_shape_error(node))

    def visit_waveform_Smooth(self, node: waveform.Smooth) -> None:
        raise ValueError(self.emit_shape_error(node))

    def visit_waveform_PythonFn(self, node: waveform.PythonFn) -> None:
        raise ValueError(self.emit_shape_error(node))

    def visit_sequence_Sequence(self, node: sequence.Sequence) -> None:
        self.visit(node.pulses[self.level_coupling])

    def visit_pulse_Pulse(self, node: pulse.Pulse) -> None:
        self.visit(node.fields[self.field_name])

    def visit_field_Field(self, node: field.Field) -> None:
        self.visit(node.drives[self.spatial_modulations])

    # node type doesn't matter here
    def scan(self, node) -> None:
        self.visit(node)
